from __future__ import annotations
from typing import Optional, List, Dict

import sqlalchemy as sql

from .expr import Expr, ExprScope, _GLOBAL_SCOPE
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog
import pixeltable.type_system as ts


class JsonMapper(Expr):
    """
    JsonMapper transforms the list output of a JsonPath by applying a target expr to every element of the list.
    The target expr would typically contain relative JsonPaths, which are bound to an ObjectRef, which in turn
    is populated by JsonMapper.eval(). The JsonMapper effectively creates a new scope for its target expr.
    """
    def __init__(self, src_expr: Expr, target_expr: Expr):
        # TODO: type spec should be List[target_expr.col_type]
        super().__init__(ts.JsonType())

        # we're creating a new scope, but we don't know yet whether this is nested within another JsonMapper;
        # this gets resolved in bind_rel_paths(); for now we assume we're in the global scope
        self.target_expr_scope = ExprScope(_GLOBAL_SCOPE)

        from .object_ref import ObjectRef
        scope_anchor = ObjectRef(self.target_expr_scope, self)
        self.components = [src_expr, target_expr, scope_anchor]
        self.parent_mapper: Optional[JsonMapper] = None
        self.target_expr_eval_ctx: Optional[RowBuilder.EvalCtx] = None
        self.id = self._create_id()

    def bind_rel_paths(self, mapper: Optional[JsonMapper]) -> None:
        self._src_expr.bind_rel_paths(mapper)
        self._target_expr.bind_rel_paths(self)
        self.parent_mapper = mapper
        parent_scope = _GLOBAL_SCOPE if mapper is None else mapper.target_expr_scope
        self.target_expr_scope.parent = parent_scope

    def scope(self) -> ExprScope:
        # need to ignore target_expr
        return self._src_expr.scope()

    def dependencies(self) -> List[Expr]:
        result = [self._src_expr]
        result.extend(self._target_dependencies(self._target_expr))
        return result

    def _target_dependencies(self, e: Expr) -> List[Expr]:
        """
        Return all subexprs of e of which the scope isn't contained in target_expr_scope.
        Those need to be evaluated before us.
        """
        expr_scope = e.scope()
        if not expr_scope.is_contained_in(self.target_expr_scope):
            return [e]
        result: List[Expr] = []
        for c in e.components:
            result.extend(self._target_dependencies(c))
        return result

    def equals(self, other: Expr) -> bool:
        """
        We override equals() because we need to avoid comparing our scope anchor.
        """
        if type(self) != type(other):
            return False
        return self._src_expr.equals(other._src_expr) and self._target_expr.equals(other._target_expr)

    def __str__(self) -> str:
        return f'{str(self._src_expr)} >> {str(self._target_expr)}'

    @property
    def _src_expr(self) -> Expr:
        return self.components[0]

    @property
    def _target_expr(self) -> Expr:
        return self.components[1]

    @property
    def scope_anchor(self) -> Expr:
        return self.components[2]

    def _equals(self, other: JsonMapper) -> bool:
        return True

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        # this will be called, but the value has already been materialized elsewhere
        src = data_row[self._src_expr.slot_idx]
        if not isinstance(src, list):
            # invalid/non-list src path
            data_row[self.slot_idx] = None
            return

        result = [None] * len(src)
        if self.target_expr_eval_ctx is None:
            self.target_expr_eval_ctx = row_builder.create_eval_ctx([self._target_expr])
        for i, val in enumerate(src):
            data_row[self.scope_anchor.slot_idx] = val
            # stored target_expr
            exc_tb = row_builder.eval(data_row, self.target_expr_eval_ctx)
            assert exc_tb is None
            result[i] = data_row[self._target_expr.slot_idx]
        data_row[self.slot_idx] = result

    def _as_dict(self) -> Dict:
        """
        We need to avoid serializing component[2], which is an ObjectRef.
        """
        return {'components': [c.as_dict() for c in self.components[0:2]]}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert len(components) == 2
        return cls(components[0], components[1])




from __future__ import annotations
from typing import List, Optional, Any, Tuple, Dict
import threading
from collections import defaultdict
from uuid import UUID
import concurrent
import logging
import urllib
from pathlib import Path

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode
import pixeltable.exprs as exprs
from pixeltable.utils.filecache import FileCache
import pixeltable.env as env
import pixeltable.exceptions as excs

_logger = logging.getLogger('pixeltable')

class CachePrefetchNode(ExecNode):
    """Brings files with external URLs into the cache

    TODO:
    - maintain a queue of row batches, in order to overlap download and evaluation
    - adapting the number of download threads at runtime to maximize throughput
    """
    def __init__(self, tbl_id: UUID, file_col_info: List[exprs.ColumnSlotIdx], input: ExecNode):
        # []: we don't have anything to evaluate
        super().__init__(input.row_builder, [], [], input)
        self.tbl_id = tbl_id
        self.file_col_info = file_col_info

        # clients for specific services are constructed as needed, because it's time-consuming
        self.boto_client: Optional[Any] = None
        self.boto_client_lock = threading.Lock()

    def __next__(self) -> DataRowBatch:
        input_batch = next(self.input)

        # collect external URLs that aren't already cached, and set DataRow.file_paths for those that are
        file_cache = FileCache.get()
        cache_misses: List[Tuple[exprs.DataRow, exprs.ColumnSlotIdx]] = []
        missing_url_rows: Dict[str, List[exprs.DataRow]] = defaultdict(list)  # URL -> rows in which it's missing
        for row in input_batch:
            for info in self.file_col_info:
                url = row.file_urls[info.slot_idx]
                if url is None or row.file_paths[info.slot_idx] is not None:
                    # nothing to do
                    continue
                if url in missing_url_rows:
                    missing_url_rows[url].append(row)
                    continue
                local_path = file_cache.lookup(url)
                if local_path is None:
                    cache_misses.append((row, info))
                    missing_url_rows[url].append(row)
                else:
                    row.set_file_path(info.slot_idx, str(local_path))

        # download the cache misses in parallel
        # TODO: set max_workers to maximize throughput
        futures: Dict[concurrent.futures.Future, Tuple[exprs.DataRow, exprs.ColumnSlotIdx]] = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=16) as executor:
            for row, info in cache_misses:
                futures[executor.submit(self._fetch_url, row, info.slot_idx)] = (row, info)
            for future in concurrent.futures.as_completed(futures):
                # TODO:  does this need to deal with recoverable errors (such as retry after throttling)?
                tmp_path = future.result()
                if tmp_path is None:
                    continue
                row, info = futures[future]
                url = row.file_urls[info.slot_idx]
                local_path = file_cache.add(self.tbl_id, info.col.id, url, tmp_path)
                _logger.debug(f'PrefetchNode: cached {url} as {local_path}')
                for row in missing_url_rows[url]:
                    row.set_file_path(info.slot_idx, str(local_path))

        return input_batch

    def _fetch_url(self, row: exprs.DataRow, slot_idx: int) -> Optional[str]:
        """Fetches a remote URL into Env.tmp_dir and returns its path"""
        url = row.file_urls[slot_idx]
        parsed = urllib.parse.urlparse(url)
        assert parsed.scheme != '' and parsed.scheme != 'file'
        # preserve the file extension, if there is one
        extension = ''
        if parsed.path != '':
            p = Path(urllib.parse.unquote(parsed.path))
            extension = p.suffix
        tmp_path = env.Env.get().create_tmp_path(extension=extension)
        try:
            if parsed.scheme == 's3':
                from pixeltable.utils.s3 import get_client
                with self.boto_client_lock:
                    if self.boto_client is None:
                        self.boto_client = get_client()
                    self.boto_client.download_file(parsed.netloc, parsed.path.lstrip('/'), str(tmp_path))
                    return tmp_path
            elif parsed.scheme == 'http' or parsed.scheme == 'https':
                with urllib.request.urlopen(url) as resp, open(tmp_path, 'wb') as f:
                    data = resp.read()
                    f.write(data)
            else:
                assert False, f'Unsupported URL scheme: {parsed.scheme}'
            return tmp_path
        except Exception as e:
            # we want to add the file url to the exception message
            exc = excs.Error(f'Failed to download {url}: {e}')
            self.row_builder.set_exc(row, slot_idx, exc)
            if not self.ctx.ignore_errors:
                raise exc from None  # suppress original exception
        return None




import sys
import warnings
from typing import List, Optional, Tuple
from dataclasses import dataclass, field
import logging
import time

from tqdm import tqdm, TqdmWarning

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode
import pixeltable.exprs as exprs
import pixeltable.func as func


_logger = logging.getLogger('pixeltable')

class ExprEvalNode(ExecNode):
    """Materializes expressions
    """
    @dataclass
    class Cohort:
        """List of exprs that form an evaluation context and contain calls to at most one external function"""
        exprs: List[exprs.Expr]
        ext_function: Optional[func.BatchedFunction]
        segment_ctxs: List[exprs.RowBuilder.EvalCtx]
        target_slot_idxs: List[int]
        batch_size: int = 8

    def __init__(
            self, row_builder: exprs.RowBuilder, output_exprs: List[exprs.Expr], input_exprs: List[exprs.Expr],
            input: ExecNode
    ):
        super().__init__(row_builder, output_exprs, input_exprs, input)
        self.input_exprs = input_exprs
        input_slot_idxs = {e.slot_idx for e in input_exprs}
        # we're only materializing exprs that are not already in the input
        self.target_exprs = [e for e in output_exprs if e.slot_idx not in input_slot_idxs]
        self.pbar: Optional[tqdm] = None
        self.cohorts: List[List[ExprEvalNode.Cohort]] = []
        self._create_cohorts()

    def __next__(self) -> DataRowBatch:
        input_batch = next(self.input)
        # compute target exprs
        for cohort in self.cohorts:
            self._exec_cohort(cohort, input_batch)
        _logger.debug(f'ExprEvalNode: returning {len(input_batch)} rows')
        return input_batch

    def _open(self) -> None:
        warnings.simplefilter("ignore", category=TqdmWarning)
        if self.ctx.show_pbar:
            self.pbar = tqdm(
                total=len(self.target_exprs) * self.ctx.num_rows,
                desc='Computing cells',
                unit=' cells',
                ncols=100,
                file=sys.stdout
            )

    def _close(self) -> None:
        if self.pbar is not None:
            self.pbar.close()

    def _get_batched_fn(self, expr: exprs.Expr) -> Optional[func.BatchedFunction]:
        if not isinstance(expr, exprs.FunctionCall):
            return None
        return expr.fn if isinstance(expr.fn, func.BatchedFunction) else None

    def _is_ext_call(self, expr: exprs.Expr) -> bool:
        return self._get_batched_fn(expr) is not None

    def _create_cohorts(self) -> None:
        all_exprs = self.row_builder.get_dependencies(self.target_exprs)
        # break up all_exprs into cohorts such that each cohort contains calls to at most one external function;
        # seed the cohorts with only the ext fn calls
        cohorts: List[List[exprs.Expr]] = []
        current_ext_function: Optional[func.BatchedFunction] = None
        for e in all_exprs:
            if not self._is_ext_call(e):
                continue
            if current_ext_function is None or current_ext_function != e.fn:
                # create a new cohort
                cohorts.append([])
                current_ext_function = e.fn
            cohorts[-1].append(e)

        # expand the cohorts to include all exprs that are in the same evaluation context as the external calls;
        # cohorts are evaluated in order, so we can exclude the target slots from preceding cohorts and input slots
        exclude = set([e.slot_idx for e in self.input_exprs])
        all_target_slot_idxs = set([e.slot_idx for e in self.target_exprs])
        target_slot_idxs: List[List[int]] = []  # the ones materialized by each cohort
        for i in range(len(cohorts)):
            cohorts[i] = self.row_builder.get_dependencies(
                cohorts[i], exclude=[self.row_builder.unique_exprs[slot_idx] for slot_idx in exclude])
            target_slot_idxs.append(
                [e.slot_idx for e in cohorts[i] if e.slot_idx in all_target_slot_idxs])
            exclude.update(target_slot_idxs[-1])

        all_cohort_slot_idxs = set([e.slot_idx for cohort in cohorts for e in cohort])
        remaining_slot_idxs = set(all_target_slot_idxs) - all_cohort_slot_idxs
        if len(remaining_slot_idxs) > 0:
            cohorts.append(self.row_builder.get_dependencies(
                [self.row_builder.unique_exprs[slot_idx] for slot_idx in remaining_slot_idxs],
                exclude=[self.row_builder.unique_exprs[slot_idx] for slot_idx in exclude]))
            target_slot_idxs.append(list(remaining_slot_idxs))
        # we need to have captured all target slots at this point
        assert all_target_slot_idxs == set().union(*target_slot_idxs)

        for i in range(len(cohorts)):
            cohort = cohorts[i]
            # segment the cohort into sublists that contain either a single ext. function call or no ext. function calls
            # (i.e., only computed cols)
            assert len(cohort) > 0
            # create the first segment here, so we can avoid checking for an empty list in the loop
            segments = [[cohort[0]]]
            is_ext_segment = self._is_ext_call(cohort[0])
            ext_fn: Optional[func.BatchedFunction] = self._get_batched_fn(cohort[0])
            for e in cohort[1:]:
                if self._is_ext_call(e):
                    segments.append([e])
                    is_ext_segment = True
                    ext_fn = self._get_batched_fn(e)
                else:
                    if is_ext_segment:
                        # start a new segment
                        segments.append([])
                        is_ext_segment = False
                    segments[-1].append(e)

            # we create the EvalCtxs manually because create_eval_ctx() would repeat the dependencies of each segment
            segment_ctxs = [
                exprs.RowBuilder.EvalCtx(
                    slot_idxs=[e.slot_idx for e in s], exprs=s, target_slot_idxs=[], target_exprs=[])
                for s in segments
            ]
            cohort_info = self.Cohort(cohort, ext_fn, segment_ctxs, target_slot_idxs[i])
            self.cohorts.append(cohort_info)

    def _exec_cohort(self, cohort: Cohort, rows: DataRowBatch) -> None:
        """Compute the cohort for the entire input batch by dividing it up into sub-batches"""
        batch_start_idx = 0  # start row of the current sub-batch
        # for multi-resolution models, we re-assess the correct ext fn batch size for each input batch
        ext_batch_size = cohort.ext_function.get_batch_size() if cohort.ext_function is not None else None
        if ext_batch_size is not None:
            cohort.batch_size = ext_batch_size

        while batch_start_idx < len(rows):
            num_batch_rows = min(cohort.batch_size, len(rows) - batch_start_idx)
            for segment_ctx in cohort.segment_ctxs:
                if not self._is_ext_call(segment_ctx.exprs[0]):
                    # compute batch row-wise
                    for row_idx in range(batch_start_idx, batch_start_idx + num_batch_rows):
                        self.row_builder.eval(
                            rows[row_idx], segment_ctx, self.ctx.profile, ignore_errors=self.ctx.ignore_errors)
                else:
                    fn_call = segment_ctx.exprs[0]
                    # make a batched external function call
                    arg_batches = [[] for _ in range(len(fn_call.args))]
                    kwarg_batches = {k: [] for k in fn_call.kwargs.keys()}

                    valid_batch_idxs: List[int] = []  # rows with exceptions are not valid
                    for row_idx in range(batch_start_idx, batch_start_idx + num_batch_rows):
                        row = rows[row_idx]
                        if row.has_exc(fn_call.slot_idx):
                            # one of our inputs had an exception, skip this row
                            continue
                        valid_batch_idxs.append(row_idx)
                        args, kwargs = fn_call._make_args(row)
                        [arg_batches[i].append(args[i]) for i in range(len(args))]
                        [kwarg_batches[k].append(kwargs[k]) for k in kwargs.keys()]
                    num_valid_batch_rows = len(valid_batch_idxs)

                    if ext_batch_size is None:
                        # we need to choose a batch size based on the args
                        sample_args = [arg_batches[i][0] for i in range(len(arg_batches))]
                        ext_batch_size = fn_call.fn.get_batch_size(*sample_args)

                    num_remaining_batch_rows = num_valid_batch_rows
                    while num_remaining_batch_rows > 0:
                        # we make ext. fn calls in batches of ext_batch_size
                        if ext_batch_size is None:
                            pass
                        num_ext_batch_rows = min(ext_batch_size, num_remaining_batch_rows)
                        ext_batch_offset = num_valid_batch_rows - num_remaining_batch_rows  # offset into args, not rows
                        call_args = [
                            arg_batches[i][ext_batch_offset:ext_batch_offset + num_ext_batch_rows]
                            for i in range(len(arg_batches))
                        ]
                        call_kwargs = {
                            k: kwarg_batches[k][ext_batch_offset:ext_batch_offset + num_ext_batch_rows]
                            for k in kwarg_batches.keys()
                        }
                        start_ts = time.perf_counter()
                        result_batch = fn_call.fn.invoke(call_args, call_kwargs)
                        self.ctx.profile.eval_time[fn_call.slot_idx] += time.perf_counter() - start_ts
                        self.ctx.profile.eval_count[fn_call.slot_idx] += num_ext_batch_rows

                        # move the result into the row batch
                        for result_idx in range(len(result_batch)):
                            row_idx = valid_batch_idxs[ext_batch_offset + result_idx]
                            row = rows[row_idx]
                            row[fn_call.slot_idx] = result_batch[result_idx]

                        num_remaining_batch_rows -= num_ext_batch_rows

                    # switch to the ext fn batch size
                    cohort.batch_size = ext_batch_size

            # make sure images for stored cols have been saved to files before moving on to the next batch
            rows.flush_imgs(
                slice(batch_start_idx, batch_start_idx + num_batch_rows), self.stored_img_cols, self.flushed_img_slots)
            if self.pbar is not None:
                self.pbar.update(num_batch_rows * len(cohort.target_slot_idxs))
            batch_start_idx += num_batch_rows




from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
from uuid import UUID
import dataclasses
import logging

import sqlalchemy as sql
import sqlalchemy.orm as orm

from .table_version import TableVersion
from .table_version_path import TableVersionPath
from .table import Table
from .named_function import NamedFunction
from .path_dict import PathDict
import pixeltable.env as env
import pixeltable.metadata.schema as schema

_logger = logging.getLogger('pixeltable')

class Catalog:
    """A repository of catalog objects"""
    _instance: Optional[Catalog] = None

    @classmethod
    def get(cls) -> Catalog:
        if cls._instance is None:
            cls._instance = cls()
            with orm.Session(env.Env.get().engine, future=True) as session:
                cls._instance._load_table_versions(session)
                #cls._instance._load_functions(session)
        return cls._instance

    @classmethod
    def clear(cls) -> None:
        """Remove the instance. Used for testing."""
        cls._instance = None

    def __init__(self) -> None:
        # key: [id, version]
        # - mutable version of a table: version == None (even though TableVersion.version is set correctly)
        # - snapshot versions: records the version of the snapshot
        self.tbl_versions: Dict[Tuple[UUID, int], TableVersion] = {}

        self.tbls: Dict[UUID, Table] = {}  # don't use a defaultdict here, it doesn't cooperate with the debugger
        self.tbl_dependents: Dict[UUID, List[Table]] = {}

        self._init_store()
        self.paths = PathDict()  # do this after _init_catalog()

    def _init_store(self) -> None:
        """One-time initialization of the stored catalog. Idempotent."""
        with orm.Session(env.Env.get().engine, future=True) as session:
            if session.query(sql.func.count(schema.Dir.id)).scalar() > 0:
                return
            # create a top-level directory, so that every schema object has a directory
            dir_md = schema.DirMd(name='')
            dir_record = schema.Dir(parent_id=None, md=dataclasses.asdict(dir_md))
            session.add(dir_record)
            session.flush()
            session.commit()
            _logger.info(f'Initialized catalog')

    def _load_snapshot_version(
            self, tbl_id: UUID, version: int, base: Optional[TableVersion], session: orm.Session
    ) -> TableVersion:
        q = session.query(schema.Table, schema.TableSchemaVersion) \
            .select_from(schema.Table) \
            .join(schema.TableVersion) \
            .join(schema.TableSchemaVersion) \
            .where(schema.Table.id == tbl_id) \
            .where(sql.text(f"({schema.TableVersion.__table__}.md->>'version')::int = {version}")) \
            .where(sql.text((
                f"({schema.TableVersion.__table__}.md->>'schema_version')::int = "
                f"{schema.TableSchemaVersion.__table__}.{schema.TableSchemaVersion.schema_version.name}")))
        tbl_record, schema_version_record = q.one()
        tbl_md = schema.md_from_dict(schema.TableMd, tbl_record.md)
        schema_version_md = schema.md_from_dict(schema.TableSchemaVersionMd, schema_version_record.md)
        # we ignore tbl_record.base_tbl_id/base_snapshot_id and use 'base' instead: if the base is a snapshot
        # we'd have to look that up first
        return TableVersion(tbl_record.id, tbl_md, version, schema_version_md, is_snapshot=True, base=base)

    def _load_table_versions(self, session: orm.Session) -> None:
        from .insertable_table import InsertableTable
        from .view import View

        # load tables/views;
        # do this in ascending order of creation ts so that we can resolve base references in one pass
        q = session.query(schema.Table, schema.TableSchemaVersion) \
            .select_from(schema.Table) \
            .join(schema.TableVersion) \
            .join(schema.TableSchemaVersion) \
            .where(sql.text(f"({schema.TableVersion.__table__}.md->>'version')::int = 0")) \
            .where(sql.text((
                f"({schema.Table.__table__}.md->>'current_schema_version')::int = "
                f"{schema.TableSchemaVersion.__table__}.{schema.TableSchemaVersion.schema_version.name}"))) \
            .order_by(sql.text(f"({schema.TableVersion.__table__}.md->>'created_at')::float"))

        for tbl_record, schema_version_record in q.all():
            tbl_md = schema.md_from_dict(schema.TableMd, tbl_record.md)
            schema_version_md = schema.md_from_dict(schema.TableSchemaVersionMd, schema_version_record.md)
            view_md = tbl_md.view_md

            if view_md is not None:
                assert len(view_md.base_versions) > 0
                # construct a TableVersionPath for the view
                refd_versions = [(UUID(tbl_id), version) for tbl_id, version in view_md.base_versions]
                base_path: Optional[TableVersionPath] = None
                base: Optional[TableVersion] = None
                # go through the versions in reverse order, so we can construct TableVersionPaths
                for base_id, version in refd_versions[::-1]:
                    base_version = self.tbl_versions.get((base_id, version), None)
                    if base_version is None:
                        if version is None:
                            # debugging
                            pass
                        # if this is a reference to a mutable table, we should have loaded it already
                        assert version is not None
                        base_version = self._load_snapshot_version(base_id, version, base, session)
                    base_path = TableVersionPath(base_version, base=base_path)
                    base = base_version
                assert base_path is not None

                base_tbl = self.tbls[base_path.tbl_version.id]
                is_snapshot = view_md is not None and view_md.is_snapshot
                snapshot_only = is_snapshot and view_md.predicate is None and len(schema_version_md.columns) == 0
                if snapshot_only:
                    # this is a pure snapshot, without a physical table backing it
                    view_path = base_path
                else:
                    tbl_version = TableVersion(
                        tbl_record.id, tbl_md, tbl_md.current_version, schema_version_md, is_snapshot=is_snapshot,
                        base=base_path.tbl_version if is_snapshot else None,
                        base_path=base_path if not is_snapshot else None)
                    view_path = TableVersionPath(tbl_version, base=base_path)

                tbl = View(
                    tbl_record.id, tbl_record.dir_id, tbl_md.name, view_path, base_tbl,
                    snapshot_only=snapshot_only)
                self.tbl_dependents[base_tbl._id].append(tbl)

            else:
                tbl_version = TableVersion(tbl_record.id, tbl_md, tbl_md.current_version, schema_version_md)
                tbl = InsertableTable(tbl_record.dir_id, tbl_version)

            self.tbls[tbl._id] = tbl
            self.tbl_dependents[tbl._id] = []
            self.paths.add_schema_obj(tbl._dir_id, tbl_md.name, tbl)

    # def _load_functions(self, session: orm.Session) -> None:
    #     # load Function metadata; doesn't load the actual callable, which can be large and is only done on-demand by the
    #     # FunctionRegistry
    #     q = session.query(schema.Function.id, schema.Function.dir_id, schema.Function.md) \
    #         .where(sql.text(f"({schema.Function.__table__}.md->>'name')::text IS NOT NULL"))
    #     for id, dir_id, md in q.all():
    #         assert 'name' in md
    #         name = md['name']
    #         assert name is not None
    #         named_fn = NamedFunction(id, dir_id, name)
    #         self.paths.add_schema_obj(dir_id, name, named_fn)



from __future__ import annotations
import datetime
import os
from typing import Optional, Dict, Any, List
from pathlib import Path
import sqlalchemy as sql
import uuid
import importlib
import importlib.util

import http.server
import socketserver
import threading
import typing
import uuid
from pathlib import Path
from typing import Optional, Dict, Any, List

import yaml
from sqlalchemy_utils.functions import database_exists, create_database, drop_database
import pgserver
import logging
import sys
import glob

from pixeltable import metadata
import pixeltable.exceptions as excs

if typing.TYPE_CHECKING:
    import openai

class Env:
    """
    Store for runtime globals.
    """
    _instance: Optional[Env] = None
    _log_fmt_str = '%(asctime)s %(levelname)s %(name)s %(filename)s:%(lineno)d: %(message)s'

    @classmethod
    def get(cls) -> Env:
        if cls._instance is None:
            cls._instance = Env()
        return cls._instance

    def __init__(self):
        self._home: Optional[Path] = None
        self._media_dir: Optional[Path] = None  # computed media files
        self._file_cache_dir: Optional[Path] = None  # cached media files with external URL
        self._dataset_cache_dir: Optional[Path] = None  # cached datasets (eg, pytorch or COCO)
        self._log_dir: Optional[Path] = None  # log files
        self._tmp_dir: Optional[Path] = None  # any tmp files
        self._sa_engine: Optional[sql.engine.base.Engine] = None
        self._pgdata_dir : Optional[Path] = None
        self._db_name: Optional[str] = None
        self._db_server: Optional[pgserver.PostgresServer] = None
        self._db_url: Optional[str] = None

        # info about installed packages that are utilized by some parts of the code;
        # package name -> version; version == []: package is installed, but we haven't determined the version yet
        self._installed_packages: Dict[str, Optional[List[int]]] = {}
        self._nos_client: Optional[Any] = None
        self._openai_client: Optional['openai.OpenAI'] = None
        self._has_together_client: bool = False
        self._spacy_nlp: Optional[Any] = None  # spacy.Language
        self._httpd: Optional[socketserver.TCPServer] = None
        self._http_address: Optional[str] = None

        # logging-related state
        self._logger = logging.getLogger('pixeltable')
        self._logger.setLevel(logging.DEBUG)  # allow everything to pass, we filter in _log_filter()
        self._logger.propagate = False
        self._logger.addFilter(self._log_filter)
        self._default_log_level = logging.INFO
        self._logfilename: Optional[str] = None
        self._log_to_stdout = False
        self._module_log_level: Dict[str, int] = {}  # module name -> log level

        # config
        self._config_file: Optional[Path] = None
        self._config: Optional[Dict[str, Any]] = None

        # create logging handler to also log to stdout
        self._stdout_handler = logging.StreamHandler(stream=sys.stdout)
        self._stdout_handler.setFormatter(logging.Formatter(self._log_fmt_str))
        self._initialized = False

    @property
    def config(self):
        return self._config

    @property
    def db_url(self) -> str:
        assert self._db_url is not None
        return self._db_url
    
    @property
    def http_address(self) -> str:
        assert self._http_address is not None
        return self._http_address

    def print_log_config(self) -> None:
        print(f'logging to {self._logfilename}')
        print(f'{"" if self._log_to_stdout else "not "}logging to stdout')
        print(f'default log level: {logging.getLevelName(self._default_log_level)}')
        print(
            f'module log levels: '
            f'{",".join([name + ":" + logging.getLevelName(val) for name, val in self._module_log_level.items()])}')

    def log_to_stdout(self, enable: bool = True) -> None:
        self._log_to_stdout = enable
        if enable:
            self._logger.addHandler(self._stdout_handler)
        else:
            self._logger.removeHandler(self._stdout_handler)

    def set_log_level(self, level: int) -> None:
        self._default_log_level = level

    def set_module_log_level(self, module: str, level: Optional[int]) -> None:
        if level is None:
            self._module_log_level.pop(module, None)
        else:
            self._module_log_level[module] = level

    def is_installed_package(self, package_name: str) -> bool:
        return self._installed_packages[package_name] is not None

    def _log_filter(self, record: logging.LogRecord) -> bool:
        if record.name == 'pixeltable':
            # accept log messages from a configured pixeltable module (at any level of the module hierarchy)
            path_parts = list(Path(record.pathname).parts)
            path_parts.reverse()
            max_idx = path_parts.index('pixeltable')
            for module_name in path_parts[:max_idx]:
                if module_name in self._module_log_level and record.levelno >= self._module_log_level[module_name]:
                    return True
        if record.levelno >= self._default_log_level:
            return True
        else:
            return False

    def set_up(self, echo: bool = False, reinit_db: bool = False) -> None:
        if self._initialized:
            return
        
        self._initialized = True
        home = Path(os.environ.get('PIXELTABLE_HOME', str(Path.home() / '.pixeltable')))
        assert self._home is None or self._home == home
        self._home = home
        self._config_file = Path(os.environ.get('PIXELTABLE_CONFIG', str(self._home / 'config.yaml')))
        self._media_dir = self._home / 'media'
        self._file_cache_dir = self._home / 'file_cache'
        self._dataset_cache_dir = self._home / 'dataset_cache'
        self._log_dir = self._home / 'logs'
        self._tmp_dir = self._home / 'tmp'
        self._pgdata_dir = Path(os.environ.get('PIXELTABLE_PGDATA', str(self._home / 'pgdata')))

        # Read in the config
        if os.path.isfile(self._config_file):
            with open(self._config_file, 'r') as stream:
                try:
                    self._config = yaml.safe_load(stream)
                except yaml.YAMLError as exc:
                    self._logger.error(f'Could not read config file: {self._config_file}')
                    self._config = {}
        else:
            self._config = {}

        if self._home.exists() and not self._home.is_dir():
            raise RuntimeError(f'{self._home} is not a directory')

        if not self._home.exists():
            # we don't have our logger set up yet, so print to stdout
            print(f'Creating a Pixeltable instance at: {self._home}')
            self._home.mkdir()
            # TODO (aaron-siegel) This is the existing behavior, but it seems scary. If something happens to
            # self._home, it will cause the DB to be destroyed even if pgdata is in an alternate location.
            # PROPOSAL: require `reinit_db` to be set explicitly to destroy the DB.
            reinit_db = True

        if not self._media_dir.exists():
            self._media_dir.mkdir()
        if not self._file_cache_dir.exists():
            self._file_cache_dir.mkdir()
        if not self._dataset_cache_dir.exists():
            self._dataset_cache_dir.mkdir()
        if not self._log_dir.exists():
            self._log_dir.mkdir()
        if not self._tmp_dir.exists():
            self._tmp_dir.mkdir()

        # configure _logger to log to a file
        self._logfilename = datetime.datetime.now().strftime('%Y%m%d_%H%M%S') + '.log'
        fh = logging.FileHandler(self._log_dir / self._logfilename, mode='w')
        fh.setFormatter(logging.Formatter(self._log_fmt_str))
        self._logger.addHandler(fh)
        sql_logger = logging.getLogger('sqlalchemy.engine')
        sql_logger.setLevel(logging.INFO)
        sql_logger.addHandler(fh)
        sql_logger.propagate = False

        # empty tmp dir
        for path in glob.glob(f'{self._tmp_dir}/*'):
            os.remove(path)

        self._db_name = os.environ.get('PIXELTABLE_DB', 'pixeltable')

        # cleanup_mode=None will leave db on for debugging purposes
        self._db_server = pgserver.get_server(self._pgdata_dir, cleanup_mode=None)
        self._db_url = self._db_server.get_uri(database=self._db_name)

        if reinit_db:
            if database_exists(self.db_url):
                drop_database(self.db_url)

        if not database_exists(self.db_url):
            self._logger.info(f'creating database at {self.db_url}')
            create_database(self.db_url)
            self._sa_engine = sql.create_engine(self.db_url, echo=echo, future=True)
            from pixeltable.metadata import schema
            schema.Base.metadata.create_all(self._sa_engine)
            metadata.create_system_info(self._sa_engine)
            # enable pgvector
            with self._sa_engine.begin() as conn:
                conn.execute(sql.text('CREATE EXTENSION vector'))
        else:
            self._logger.info(f'found database {self.db_url}')
            if self._sa_engine is None:
                self._sa_engine = sql.create_engine(self.db_url, echo=echo, future=True)

        print(f'Connected to Pixeltable database at: {self.db_url}')

        # we now have a home directory and db; start other services
        self._set_up_runtime()
        self.log_to_stdout(False)

    def upgrade_metadata(self) -> None:
        metadata.upgrade_md(self._sa_engine)

    def _create_nos_client(self) -> None:
        import nos
        self._logger.info('connecting to NOS')
        nos.init(logging_level=logging.DEBUG)
        self._nos_client = nos.client.InferenceClient()
        self._logger.info('waiting for NOS')
        self._nos_client.WaitForServer()

        # now that we have a client, we can create the module
        import importlib
        try:
            importlib.import_module('pixeltable.functions.nos')
            # it's already been created
            return
        except ImportError:
            pass
        from pixeltable.functions.util import create_nos_modules
        _ = create_nos_modules()

    def _create_openai_client(self) -> None:
        if 'openai' in self._config and 'api_key' in self._config['openai']:
            api_key = self._config['openai']['api_key']
        else:
            api_key = os.environ.get('OPENAI_API_KEY')
        if api_key is None or api_key == '':
            self._logger.info("OpenAI client not initialized (no API key configured).")
            return
        import openai
        self._logger.info('Initializing OpenAI client.')
        self._openai_client = openai.OpenAI(api_key=api_key)

    def _create_together_client(self) -> None:
        if 'together' in self._config and 'api_key' in self._config['together']:
            api_key = self._config['together']['api_key']
        else:
            api_key = os.environ.get('TOGETHER_API_KEY')
        if api_key is None or api_key == '':
            self._logger.info('Together client not initialized (no API key configured).')
            return
        import together
        self._logger.info('Initializing Together client.')
        together.api_key = api_key
        self._has_together_client = True

    def _start_web_server(self) -> None:
        """
        The http server root is the file system root.
        eg: /home/media/foo.mp4 is located at http://127.0.0.1:{port}/home/media/foo.mp4
        This arrangement enables serving media hosted within _home, 
        as well as external media inserted into pixeltable or produced by pixeltable.
        The port is chosen dynamically to prevent conflicts.
        """        
        # Port 0 means OS picks one for us.
        address = ("127.0.0.1", 0)
        class FixedRootHandler(http.server.SimpleHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory='/', **kwargs)        
        self._httpd = socketserver.TCPServer(address, FixedRootHandler)
        port = self._httpd.server_address[1]
        self._http_address = f'http://127.0.0.1:{port}'

        def run_server():
            logging.log(logging.INFO, f'running web server at {self._http_address}')
            self._httpd.serve_forever()

        # Run the server in a separate thread
        thread = threading.Thread(target=run_server, daemon=True)
        thread.start()

    def _set_up_runtime(self) -> None:
        """Check for and start runtime services"""
        self._start_web_server()
        self._check_installed_packages()

    def _check_installed_packages(self) -> None:
        def check(package: str) -> None:
            if importlib.util.find_spec(package) is not None:
                self._installed_packages[package] = []
            else:
                self._installed_packages[package] = None

        check('torch')
        check('torchvision')
        check('transformers')
        check('sentence_transformers')
        check('boto3')
        check('pyarrow')
        check('spacy')  # TODO: deal with en-core-web-sm
        if self.is_installed_package('spacy'):
            import spacy
            self._spacy_nlp = spacy.load('en_core_web_sm')
        check('tiktoken')
        check('openai')
        if self.is_installed_package('openai'):
            self._create_openai_client()
        check('together')
        if self.is_installed_package('together'):
            self._create_together_client()
        check('fireworks')
        check('nos')
        if self.is_installed_package('nos'):
            self._create_nos_client()

    def require_package(self, package: str, min_version: Optional[List[int]] = None) -> None:
        assert package in self._installed_packages
        if self._installed_packages[package] is None:
            raise excs.Error(f'Package {package} is not installed')
        if min_version is None:
            return

        # check whether we have a version >= the required one
        if self._installed_packages[package] == []:
            m = importlib.import_module(package)
            module_version = [int(x) for x in m.__version__.split('.')]
            self._installed_packages[package] = module_version
        installed_version = self._installed_packages[package]
        if len(min_version) < len(installed_version):
            normalized_min_version = min_version + [0] * (len(installed_version) - len(min_version))
        if any([a < b for a, b in zip(installed_version, normalized_min_version)]):
            raise excs.Error((
                f'The installed version of package {package} is {".".join([str[v] for v in installed_version])}, '
                f'but version  >={".".join([str[v] for v in min_version])} is required'))

    def num_tmp_files(self) -> int:
        return len(glob.glob(f'{self._tmp_dir}/*'))

    def create_tmp_path(self, extension: str = '') -> Path:
        return self._tmp_dir / f'{uuid.uuid4()}{extension}'

    @property
    def home(self) -> Path:
        assert self._home is not None
        return self._home

    @property
    def media_dir(self) -> Path:
        assert self._media_dir is not None
        return self._media_dir

    @property
    def file_cache_dir(self) -> Path:
        assert self._file_cache_dir is not None
        return self._file_cache_dir

    @property
    def dataset_cache_dir(self) -> Path:
        assert self._dataset_cache_dir is not None
        return self._dataset_cache_dir

    @property
    def tmp_dir(self) -> Path:
        assert self._tmp_dir is not None
        return self._tmp_dir

    @property
    def engine(self) -> sql.engine.base.Engine:
        assert self._sa_engine is not None
        return self._sa_engine

    @property
    def nos_client(self) -> Any:
        return self._nos_client

    @property
    def openai_client(self) -> Optional['openai.OpenAI']:
        return self._openai_client

    @property
    def has_together_client(self) -> bool:
        return self._has_together_client

    @property
    def spacy_nlp(self) -> Any:
        assert self._spacy_nlp is not None
        return self._spacy_nlp


from __future__ import annotations

import dataclasses
import enum
import inspect
import logging
import typing
from typing import Optional, Callable, Dict, List, Any, Union, Tuple

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

_logger = logging.getLogger('pixeltable')


@dataclasses.dataclass
class Parameter:
    name: str
    col_type: Optional[ts.ColumnType]  # None for variable parameters
    kind: enum.Enum  # inspect.Parameter.kind; inspect._ParameterKind is private
    is_batched: bool = False  # True if the parameter is a batched parameter (eg, Batch[dict])


T = typing.TypeVar('T')
Batch = typing.Annotated[list[T], 'pxt-batch']


class Signature:
    """
    Represents the signature of a Pixeltable function.

    Regarding return type:
    - most functions will have a fixed return type, which is specified directly
    - some functions will have a return type that depends on the argument values;
      ex.: PIL.Image.Image.resize() returns an image with dimensions specified as a parameter
    - in the latter case, the 'return_type' field is a function that takes the bound arguments and returns the
      return type; if no bound arguments are specified, a generic return type is returned (eg, ImageType() without a
      size)
    - self.is_batched: return type is a Batch[...] type
    """
    SPECIAL_PARAM_NAMES = ['group_by', 'order_by']

    def __init__(
            self,
            return_type: Union[ts.ColumnType, Callable[[Dict[str, Any]], ts.ColumnType]],
            parameters: List[Parameter], is_batched: bool = False):
        self.return_type = return_type
        self.is_batched = is_batched
        # we rely on the ordering guarantee of dicts in Python >=3.7
        self.parameters = {p.name: p for p in parameters}
        self.parameters_by_pos = parameters.copy()
        self.constant_parameters = [p for p in parameters if not p.is_batched]
        self.batched_parameters = [p for p in parameters if p.is_batched]

    def get_return_type(self, bound_args: Optional[Dict[str, Any]] = None) -> ts.ColumnType:
        if isinstance(self.return_type, ts.ColumnType):
            return self.return_type
        return self.return_type(bound_args)

    def as_dict(self) -> Dict[str, Any]:
        result = {
            'return_type': self.get_return_type().as_dict(),
            'parameters': [
                [p.name, p.col_type.as_dict() if p.col_type is not None else None, p.kind, p.is_batched]
                for p in self.parameters.values()
            ]
        }
        return result

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Signature:
        parameters = [Parameter(p[0], ts.ColumnType.from_dict(p[1]), p[2], p[3]) for p in d['parameters']]
        return cls(ts.ColumnType.from_dict(d['return_type']), parameters)

    def __eq__(self, other: Signature) -> bool:
        if self.get_return_type() != other.get_return_type():
            return False
        if len(self.parameters) != len(other.parameters):
            return False
        # ignore the parameter name
        for param, other_param in zip(self.parameters.values(), other.parameters.values()):
            if param.col_type != other_param.col_type or param.kind != other_param.kind:
                return False
        return True

    def __str__(self) -> str:
        param_strs: List[str] = []
        for p in self.parameters.values():
            if p.kind == inspect.Parameter.VAR_POSITIONAL:
                param_strs.append(f'*{p.name}')
            elif p.kind == inspect.Parameter.VAR_KEYWORD:
                param_strs.append(f'**{p.name}')
            else:
                param_strs.append(f'{p.name}: {str(p.col_type)}')
        return f'({", ".join(param_strs)}) -> {str(self.get_return_type())}'

    @classmethod
    def _infer_type(cls, annotation: Optional[type]) -> Tuple[Optional[ts.ColumnType], Optional[bool]]:
        """Returns: (column type, is_batched) or (None, ...) if the type cannot be inferred"""
        if annotation is None:
            return (None, None)
        py_type: Optional[type] = None
        is_batched = False
        if typing.get_origin(annotation) == typing.Annotated:
            type_args = typing.get_args(annotation)
            if len(type_args) == 2 and type_args[1] == 'pxt-batch':
                # this is our Batch
                assert typing.get_origin(type_args[0]) == list
                is_batched = True
                py_type = typing.get_args(type_args[0])[0]
        if py_type is None:
            py_type = annotation
        col_type = ts.ColumnType.from_python_type(py_type)
        return (col_type, is_batched)

    @classmethod
    def create(
            cls, c: Callable,
            param_types: Optional[List[ts.ColumnType]] = None,
            return_type: Optional[Union[ts.ColumnType, Callable]] = None
    ) -> Signature:
        """Create a signature for the given Callable.
        Infer the parameter and return types, if none are specified.
        Raises an exception if the types cannot be inferred.
        """
        sig = inspect.signature(c)
        py_parameters = list(sig.parameters.values())

        # check non-var parameters for name collisions and default value compatibility
        parameters: List[Parameter] = []
        for idx, param in enumerate(py_parameters):
            if param.name in cls.SPECIAL_PARAM_NAMES:
                raise excs.Error(f"'{param.name}' is a reserved parameter name")
            if param.kind == inspect.Parameter.VAR_POSITIONAL or param.kind == inspect.Parameter.VAR_KEYWORD:
                parameters.append(Parameter(param.name, None, param.kind, False))
                continue

            if param_types is not None:
                if idx >= len(param_types):
                    raise excs.Error(f'Missing type for parameter {param.name}')
                param_type = param_types[idx]
                is_batched = False
            else:
                param_type, is_batched = cls._infer_type(param.annotation)
                if param_type is None:
                    raise excs.Error(f'Cannot infer pixeltable type for parameter {param.name}')

            # check default value compatibility
            default_val = sig.parameters[param.name].default
            if default_val != inspect.Parameter.empty and default_val is not None:
                try:
                    _ = param_type.create_literal(default_val)
                except TypeError as e:
                    raise excs.Error(f'Default value for parameter {param.name}: {str(e)}')

            parameters.append(Parameter(param.name, param_type, param.kind, is_batched))

        return_is_batched = False
        if return_type is None:
            return_type, return_is_batched = cls._infer_type(sig.return_annotation)
            if return_type is None:
                raise excs.Error('Cannot infer pixeltable return type')
        else:
            _, return_is_batched = cls._infer_type(sig.return_annotation)

        return Signature(return_type, parameters, return_is_batched)



from abc import abstractmethod
from typing import Optional
from uuid import UUID


class SchemaObject:
    """
    Base class of all addressable objects within a Db.
    Each object has an id, a name and a parent directory.
    """
    def __init__(self, obj_id: UUID, name: str, dir_id: Optional[UUID]):
        # make these private so they don't collide with column names (id and name are fairly common)
        self._id = obj_id
        self._name = name
        self._dir_id = dir_id

    def get_id(self) -> UUID:
        return self._id

    def get_name(self) -> str:
        return self._name

    @classmethod
    @abstractmethod
    def display_name(cls) -> str:
        """
        Return name displayed in error messages.
        """
        pass

    @property
    def fqn(self) -> str:
        return f'{self.parent_dir().fqn}.{self._name}'

    def move(self, new_name: str, new_dir_id: UUID) -> None:
        """Subclasses need to override this to make the change persistent"""
        self._name = new_name
        self._dir_id = new_dir_id




from typing import Dict, Any, List, Tuple
from pathlib import Path
import math
import logging

import cv2
import PIL.Image

from .base import ComponentIterator

from pixeltable.type_system import ColumnType, VideoType, ImageType, IntType, FloatType
from pixeltable.exceptions import Error


_logger = logging.getLogger('pixeltable')

class FrameIterator(ComponentIterator):
    def __init__(self, video: str, fps: float = 0.0):
        video_path = Path(video)
        assert video_path.exists() and video_path.is_file()
        self.video_path = video_path
        self.fps = fps
        self.video_reader = cv2.VideoCapture(str(video_path))
        if not self.video_reader.isOpened():
            raise Error(f'Failed to open video: {video}')
        video_fps = int(self.video_reader.get(cv2.CAP_PROP_FPS))
        if fps > video_fps:
            raise Error(f'Video {video}: requested fps ({fps}) exceeds that of the video ({video_fps})')
        self.frame_freq = int(video_fps / fps) if fps > 0 else 1
        num_video_frames = int(self.video_reader.get(cv2.CAP_PROP_FRAME_COUNT))
        if num_video_frames == 0:
            raise Error(f'Video {video}: failed to get number of frames')
        # ceil: round up to ensure we count frame 0
        self.num_frames = math.ceil(num_video_frames / self.frame_freq) if fps > 0 else num_video_frames
        _logger.debug(f'FrameIterator: path={self.video_path} fps={self.fps}')

        self.next_frame_idx = 0

    @classmethod
    def input_schema(cls) -> Dict[str, ColumnType]:
        return {
            'video': VideoType(nullable=False),
            'fps': FloatType()
        }

    @classmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> Tuple[Dict[str, ColumnType], List[str]]:
        return {
            'frame_idx': IntType(),
            'pos_msec': FloatType(),
            'pos_frame': FloatType(),
            'frame': ImageType(),
        }, ['frame']

    def __next__(self) -> Dict[str, Any]:
        while True:
            pos_msec = self.video_reader.get(cv2.CAP_PROP_POS_MSEC)
            pos_frame = self.video_reader.get(cv2.CAP_PROP_POS_FRAMES)
            status, img = self.video_reader.read()
            if not status:
                _logger.debug(f'releasing video reader for {self.video_path}')
                self.video_reader.release()
                self.video_reader = None
                raise StopIteration
            if pos_frame % self.frame_freq == 0:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                result = {
                    'frame_idx': self.next_frame_idx,
                    'pos_msec': pos_msec,
                    'pos_frame': pos_frame,
                    'frame': PIL.Image.fromarray(img),
                }
                self.next_frame_idx += 1
                # frame_freq > 1: jumping to the target frame here with video_reader.set() is far slower than just
                # skipping the unwanted frames
                return result

    def close(self) -> None:
        if self.video_reader is not None:
            self.video_reader.release()
            self.video_reader = None

    def set_pos(self, pos: int) -> None:
        """Seek to frame idx"""
        if pos == self.next_frame_idx:
            return
        _logger.debug(f'seeking to frame {pos}')
        self.video_reader.set(cv2.CAP_PROP_POS_FRAMES, pos * self.frame_freq)
        self.next_frame_idx = pos



import datetime
import math
import os
import random
from typing import List, Tuple

import PIL
import cv2
import numpy as np
import pandas as pd
import pytest

import pixeltable as pxt
import pixeltable.functions as ptf
from pixeltable import catalog
from pixeltable import exceptions as excs
from pixeltable.iterators import FrameIterator
from pixeltable.tests.utils import \
    make_tbl, create_table_data, read_data_file, get_video_files, get_audio_files, get_image_files, get_documents, \
    assert_resultset_eq
from pixeltable.type_system import \
    StringType, IntType, FloatType, TimestampType, ImageType, VideoType, JsonType, BoolType, ArrayType, AudioType, \
    DocumentType
from pixeltable.utils.filecache import FileCache
from pixeltable.utils.media_store import MediaStore


class TestTable:
    # exc for a % 10 == 0
    @pxt.udf(return_type=FloatType(), param_types=[IntType()])
    def f1(a: int) -> float:
        return a / (a % 10)

    # exception for a == None; this should not get triggered
    @pxt.udf(return_type=FloatType(), param_types=[FloatType()])
    def f2(a: float) -> float:
        return a + 1

    @pxt.expr_udf(param_types=[IntType(nullable=False)])
    def add1(a: int) -> int:
        return a + 1

    def test_create(self, test_client: pxt.Client) -> None:
        cl = test_client
        cl.create_dir('dir1')
        schema = {
            'c1': StringType(nullable=False),
            'c2': IntType(nullable=False),
            'c3': FloatType(nullable=False),
            'c4': TimestampType(nullable=False),
        }
        tbl = cl.create_table('test', schema)
        _ = cl.create_table('dir1.test', schema)

        with pytest.raises(excs.Error):
            _ = cl.create_table('1test', schema)
        with pytest.raises(excs.Error):
            _ = catalog.Column('1c', StringType())
        with pytest.raises(excs.Error):
            _ = cl.create_table('test', schema)
        with pytest.raises(excs.Error):
            _ = cl.create_table('dir2.test2', schema)

        _ = cl.list_tables()
        _ = cl.list_tables('dir1')

        with pytest.raises(excs.Error):
            _ = cl.list_tables('1dir')
        with pytest.raises(excs.Error):
            _ = cl.list_tables('dir2')

        # test loading with new client
        cl = pxt.Client(reload=True)

        tbl = cl.get_table('test')
        assert isinstance(tbl, catalog.InsertableTable)
        tbl.add_column(c5=IntType())
        tbl.drop_column('c1')
        tbl.rename_column('c2', 'c17')

        cl.move('test', 'test2')

        cl.drop_table('test2')
        cl.drop_table('dir1.test')

        with pytest.raises(excs.Error):
            cl.drop_table('test')
        with pytest.raises(excs.Error):
            cl.drop_table('dir1.test2')
        with pytest.raises(excs.Error):
            cl.drop_table('.test2')

    def test_empty_table(self, test_client: pxt.Client) -> None:
        cl = test_client
        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('empty_table', {})
        assert 'Table schema is empty' in str(exc_info.value)

    def test_table_attrs(self, test_client: pxt.Client) -> None:
        cl = test_client
        schema = {'c': StringType(nullable=False)}
        num_retained_versions = 20
        comment = "This is a table."
        tbl = cl.create_table('test_table_attrs', schema, num_retained_versions=num_retained_versions, comment=comment)
        assert tbl.num_retained_versions == num_retained_versions
        assert tbl.comment == comment
        new_num_retained_versions = 30
        new_comment = "This is an updated table."
        tbl.num_retained_versions = new_num_retained_versions
        assert tbl.num_retained_versions == new_num_retained_versions
        tbl.comment = new_comment
        assert tbl.comment == new_comment
        tbl.revert()
        assert tbl.comment == comment
        tbl.revert()
        assert tbl.num_retained_versions == num_retained_versions

    def test_image_table(self, test_client: pxt.Client) -> None:
        n_sample_rows = 20
        cl = test_client
        schema = {
            'img': ImageType(nullable=False),
            'category': StringType(nullable=False),
            'split': StringType(nullable=False),
            'img_literal': ImageType(nullable=False),
        }
        tbl = cl.create_table('test', schema)
        assert(MediaStore.count(tbl.get_id()) == 0)

        rows = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
        sample_rows = random.sample(rows, n_sample_rows)

        # add literal image data and column
        for r in rows:
            with open(r['img'], 'rb') as f:
                r['img_literal'] = f.read()

        tbl.insert(sample_rows)
        assert(MediaStore.count(tbl.get_id()) == n_sample_rows)

        # compare img and img_literal
        # TODO: make tbl.select(tbl.img == tbl.img_literal) work
        tdf = tbl.select(tbl.img, tbl.img_literal).show()
        pdf = tdf.to_pandas()
        for tup in pdf.itertuples():
            assert tup.img == tup.img_literal

        # Test adding stored image transformation
        tbl.add_column(rotated=tbl.img.rotate(30), stored=True)
        assert(MediaStore.count(tbl.get_id()) == 2 * n_sample_rows)

        # Test MediaStore.stats()
        stats = list(filter(lambda x: x[0] == tbl.get_id(), MediaStore.stats()))
        assert len(stats) == 2                 # Two columns
        assert stats[0][2] == n_sample_rows    # Each column has n_sample_rows associated images
        assert stats[1][2] == n_sample_rows

        # Test that version-specific images are cleared when table is reverted
        tbl.revert()
        assert(MediaStore.count(tbl.get_id()) == n_sample_rows)

        # Test that all stored images are cleared when table is dropped
        cl.drop_table('test')
        assert(MediaStore.count(tbl.get_id()) == 0)

    def test_schema_spec(self, test_client: pxt.Client) -> None:
        cl = test_client

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c 1': IntType()})
        assert 'invalid column name' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {}})
        assert '"type" is required' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'xyz': IntType()}})
        assert "invalid key 'xyz'" in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'stored': True}})
        assert '"type" is required' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'type': 'string'}})
        assert 'must be a ColumnType' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'value': 1, 'type': StringType()}})
        assert '"type" is redundant' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'value': pytest}})
        assert 'value needs to be either' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            def f() -> float:
                return 1.0
            cl.create_table('test', {'c1': {'value': f}})
        assert '"type" is required' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'type': StringType(), 'stored': 'true'}})
        assert '"stored" must be a bool' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': {'type': StringType(), 'indexed': 'true'}})
        assert '"indexed" must be a bool' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': StringType()}, primary_key='c2')
        assert 'primary key column c2 not found' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': StringType()}, primary_key=['c1', 'c2'])
        assert 'primary key column c2 not found' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': StringType()}, primary_key=['c2'])
        assert 'primary key column c2 not found' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': StringType()}, primary_key=0)
        assert 'primary_key must be a' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            cl.create_table('test', {'c1': StringType(nullable=True)}, primary_key='c1')
        assert 'cannot be nullable' in str(exc_info.value).lower()

    def check_bad_media(
            self, test_client: pxt.Client, rows: List[Tuple[str, bool]], col_type: pxt.ColumnType,
            validate_local_path: bool = True
    ) -> None:
        schema = {
            'media': col_type,
            'is_bad_media': BoolType(nullable=False),
        }
        tbl = test_client.create_table('test', schema)

        assert len(rows) > 0
        total_bad_rows = sum([int(row['is_bad_media']) for row in rows])
        assert total_bad_rows > 0

        # Mode 1: Validation error on bad input (default)
        # we ignore the exact error here, because it depends on the media type
        with pytest.raises(excs.Error):
            tbl.insert(rows, fail_on_exception=True)

        # Mode 2: ignore_errors=True, store error information in table
        status = tbl.insert(rows, fail_on_exception=False)
        _ = tbl.select(tbl.media, tbl.media.errormsg).show()
        assert status.num_rows == len(rows)
        assert status.num_excs == total_bad_rows

        # check that we have the right number of bad and good rows
        assert tbl.where(tbl.is_bad_media == True).count() == total_bad_rows
        assert tbl.where(tbl.is_bad_media == False).count() == len(rows) - total_bad_rows

        # check error type is set correctly
        assert tbl.where((tbl.is_bad_media == True) & (tbl.media.errortype == None)).count() == 0
        assert tbl.where((tbl.is_bad_media == False) & (tbl.media.errortype == None)).count() \
            == len(rows) - total_bad_rows

        # check fileurl is set for valid images, and check no file url is set for bad images
        assert tbl.where((tbl.is_bad_media == False) & (tbl.media.fileurl == None)).count() == 0
        assert tbl.where((tbl.is_bad_media == True) & (tbl.media.fileurl != None)).count() == 0

        if validate_local_path:
            # check that tbl.media is a valid local path
            paths = tbl.where(tbl.media != None).select(output=tbl.media).collect()['output']
            for path in paths:
                assert os.path.exists(path) and os.path.isfile(path)

    def test_validate_image(self, test_client: pxt.Client) -> None:
        rows = read_data_file('imagenette2-160', 'manifest_bad.csv', ['img'])
        rows = [{'media': r['img'], 'is_bad_media': r['is_bad_image']} for r in rows]
        self.check_bad_media(test_client, rows, ImageType(nullable=True), validate_local_path=False)

    def test_validate_video(self, test_client: pxt.Client) -> None:
        files = get_video_files(include_bad_video=True)
        rows = [{'media': f, 'is_bad_media': f.endswith('bad_video.mp4')} for f in files]
        self.check_bad_media(test_client, rows, VideoType(nullable=True))

    def test_validate_audio(self, test_client: pxt.Client) -> None:
        files = get_audio_files(include_bad_audio=True)
        rows = [{'media': f, 'is_bad_media': f.endswith('bad_audio.mp3')} for f in files]
        self.check_bad_media(test_client, rows, AudioType(nullable=True))

    def test_validate_docs(self, test_client: pxt.Client) -> None:
        valid_doc_paths = get_documents()
        invalid_doc_paths = [get_video_files()[0], get_audio_files()[0], get_image_files()[0]]
        doc_paths = valid_doc_paths + invalid_doc_paths
        is_valid = [True] * len(valid_doc_paths) + [False] * len(invalid_doc_paths)
        rows = [{'media': f, 'is_bad_media': not is_valid} for f, is_valid in zip(doc_paths, is_valid)]
        self.check_bad_media(test_client, rows, DocumentType(nullable=True))

    def test_validate_external_url(self, test_client: pxt.Client) -> None:
        rows = [
            {'media': 's3://open-images-dataset/validation/doesnotexist.jpg', 'is_bad_media': True},
            {'media': 'https://archive.random.org/download?file=2024-01-28.bin', 'is_bad_media': True},  # 403 error
            {'media': 's3://open-images-dataset/validation/3c02ca9ec9b2b77b.jpg', 'is_bad_media': True},  # wrong media
            # test s3 url
            {
                'media': 's3://multimedia-commons/data/videos/mp4/ffe/ff3/ffeff3c6bf57504e7a6cecaff6aefbc9.mp4',
                'is_bad_media': False
            },
            # test http url
            {
                'media': 'https://github.com/pixeltable/pixeltable/raw/master/pixeltable/tests/data/videos/bangkok.mp4',
                'is_bad_media': False
            },

        ]
        self.check_bad_media(test_client, rows, VideoType(nullable=True))

    def test_create_s3_image_table(self, test_client: pxt.Client) -> None:
        cl = test_client
        tbl = cl.create_table('test', {'img': ImageType(nullable=False)})
        # this is needed because Client.reset_catalog() doesn't call TableVersion.drop(), which would
        # clear the file cache
        # TODO: change reset_catalog() to drop tables
        FileCache.get().clear()
        cache_stats = FileCache.get().stats()
        assert cache_stats.num_requests == 0, f'{str(cache_stats)} tbl_id={tbl.get_id()}'
        # add computed column to make sure that external files are cached locally during insert
        tbl.add_column(rotated=tbl.img.rotate(30), stored=True)
        urls = [
            's3://open-images-dataset/validation/3c02ca9ec9b2b77b.jpg',
            's3://open-images-dataset/validation/3c13e0015b6c3bcf.jpg',
            's3://open-images-dataset/validation/3ba5380490084697.jpg',
            's3://open-images-dataset/validation/3afeb4b34f90c0cf.jpg',
            's3://open-images-dataset/validation/3b07a2c0d5c0c789.jpg',
        ]

        tbl.insert({'img': url} for url in urls)
        # check that we populated the cache
        cache_stats = FileCache.get().stats()
        assert cache_stats.num_requests == len(urls), f'{str(cache_stats)} tbl_id={tbl.get_id()}'
        assert cache_stats.num_hits == 0
        assert FileCache.get().num_files() == len(urls)
        assert FileCache.get().num_files(tbl.get_id()) == len(urls)
        assert FileCache.get().avg_file_size() > 0

        # query: we read from the cache
        _ = tbl.show(0)
        cache_stats = FileCache.get().stats()
        assert cache_stats.num_requests == 2 * len(urls)
        assert cache_stats.num_hits == len(urls)

        # after clearing the cache, we need to re-fetch the files
        FileCache.get().clear()
        _ = tbl.show(0)
        cache_stats = FileCache.get().stats()
        assert cache_stats.num_requests == len(urls)
        assert cache_stats.num_hits == 0

        # start with fresh client and FileCache instance to test FileCache initialization with pre-existing files
        cl = pxt.Client(reload=True)
        # is there a better way to do this?
        FileCache._instance = None
        t = cl.get_table('test')
        _ = t.show(0)
        cache_stats = FileCache.get().stats()
        assert cache_stats.num_requests == len(urls)
        assert cache_stats.num_hits == len(urls)

        # dropping the table also clears the file cache
        cl.drop_table('test')
        cache_stats = FileCache.get().stats()
        assert cache_stats.total_size == 0

    def test_video_url(self, test_client: pxt.Client) -> None:
        cl = test_client
        schema = {
            'payload': IntType(nullable=False),
            'video': VideoType(nullable=False),
        }
        tbl = cl.create_table('test', schema)
        url = 's3://multimedia-commons/data/videos/mp4/ffe/ff3/ffeff3c6bf57504e7a6cecaff6aefbc9.mp4'
        tbl.insert(payload=1, video=url)
        row = tbl.select(tbl.video.fileurl, tbl.video.localpath).collect()[0]
        assert row['video_fileurl'] == url
        # row[1] contains valid path to an mp4 file
        local_path = row['video_localpath']
        assert os.path.exists(local_path) and os.path.isfile(local_path)
        cap = cv2.VideoCapture(local_path)
        # TODO: this isn't sufficient to determine that this is actually a video, rather than an image
        assert cap.isOpened()
        cap.release()

    def test_create_video_table(self, test_client: pxt.Client) -> None:
        cl = test_client
        tbl = cl.create_table(
            'test_tbl',
            {'payload': IntType(nullable=False), 'video': VideoType(nullable=True)})
        args = {'video': tbl.video, 'fps': 0}
        view = cl.create_view('test_view', tbl, iterator_class=FrameIterator, iterator_args=args)
        view.add_column(c1=view.frame.rotate(30), stored=True)
        view.add_column(c2=view.c1.rotate(40), stored=False)
        view.add_column(c3=view.c2.rotate(50), stored=True)
        # a non-materialized column that refers to another non-materialized column
        view.add_column(c4=view.c2.rotate(60), stored=False)

        @pxt.uda(
            name='window_fn', update_types=[IntType()], value_type=IntType(), requires_order_by = True,
            allows_window = True)
        class WindowFnAggregator:
            def __init__(self):
                pass
            def update(self, i: int) -> None:
                pass
            def value(self) -> int:
                return 1
        # cols computed with window functions are stored by default
        view.add_column(c5=window_fn(view.frame_idx, 1, group_by=view.video))

        # reload to make sure that metadata gets restored correctly
        cl = pxt.Client(reload=True)
        tbl = cl.get_table('test_tbl')
        view = cl.get_table('test_view')
        # we're inserting only a single row and the video column is not in position 0
        url = 's3://multimedia-commons/data/videos/mp4/ffe/ff3/ffeff3c6bf57504e7a6cecaff6aefbc9.mp4'
        status = tbl.insert(payload=1, video=url)
        assert status.num_excs == 0
        # * 2: we have 2 stored img cols
        assert MediaStore.count(view.get_id()) == view.count() * 2
        # also insert a local file
        tbl.insert(payload=1, video=get_video_files()[0])
        assert MediaStore.count(view.get_id()) == view.count() * 2

        # TODO: test inserting Nulls
        #status = tbl.insert(payload=1, video=None)
        #assert status.num_excs == 0

        # revert() clears stored images
        tbl.revert()
        tbl.revert()
        assert MediaStore.count(view.get_id()) == 0

        with pytest.raises(excs.Error):
            # can't drop frame col
            view.drop_column('frame')
        with pytest.raises(excs.Error):
            # can't drop frame_idx col
            view.drop_column('frame_idx')

        # drop() clears stored images and the cache
        tbl.insert(payload=1, video=get_video_files()[0])
        with pytest.raises(excs.Error) as exc_info:
            cl.drop_table('test_tbl')
        assert 'has dependents: test_view' in str(exc_info.value)
        cl.drop_table('test_view')
        cl.drop_table('test_tbl')
        assert MediaStore.count(view.get_id()) == 0

    def test_insert(self, test_client: pxt.Client) -> None:
        cl = test_client
        schema = {
            'c1': StringType(nullable=False),
            'c2': IntType(nullable=False),
            'c3': FloatType(nullable=False),
            'c4': BoolType(nullable=False),
            'c5': ArrayType((2, 3), dtype=IntType(), nullable=False),
            'c6': JsonType(nullable=False),
            'c7': ImageType(nullable=False),
            'c8': VideoType(nullable=False),
        }
        t = cl.create_table('test1', schema)
        rows = create_table_data(t)
        status = t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0

        # alternate (kwargs) insert syntax
        status = t.insert(
            c1='string',
            c2=91,
            c3=1.0,
            c4=True,
            c5=np.ones((2, 3), dtype=np.dtype(np.int64)),
            c6={'key': 'val'},
            c7=get_image_files()[0],
            c8=get_video_files()[0]
        )
        assert status.num_rows == 1
        assert status.num_excs == 0

        # empty input
        with pytest.raises(excs.Error) as exc_info:
            t.insert([])
        assert 'empty' in str(exc_info.value)

        # missing column
        with pytest.raises(excs.Error) as exc_info:
            # drop first column
            col_names = list(rows[0].keys())[1:]
            new_rows = [{col_name: row[col_name] for col_name in col_names} for row in rows]
            t.insert(new_rows)
        assert 'Missing' in str(exc_info.value)

        # incompatible schema
        for (col_name, col_type), value_col_name in zip(schema.items(), ['c2', 'c3', 'c5', 'c5', 'c6', 'c7', 'c2', 'c2']):
            cl.drop_table('test1', ignore_errors=True)
            t = cl.create_table('test1', {col_name: col_type})
            with pytest.raises(excs.Error) as exc_info:
                t.insert({col_name: r[value_col_name]} for r in rows)
            assert 'expected' in str(exc_info.value).lower()

        # rows not list of dicts
        cl.drop_table('test1', ignore_errors=True)
        t = cl.create_table('test1', {'c1': StringType()})
        with pytest.raises(excs.Error) as exc_info:
            t.insert(['1'])
        assert 'list of dictionaries' in str(exc_info.value)

        # bad null value
        cl.drop_table('test1', ignore_errors=True)
        t = cl.create_table('test1', {'c1': StringType(nullable=False)})
        with pytest.raises(excs.Error) as exc_info:
            t.insert(c1=None)
        assert 'expected non-None' in str(exc_info.value)

        # bad array literal
        cl.drop_table('test1', ignore_errors=True)
        t = cl.create_table('test1', {'c5': ArrayType((2, 3), dtype=IntType(), nullable=False)})
        with pytest.raises(excs.Error) as exc_info:
            t.insert(c5=np.ndarray((3, 2)))
        assert 'expected ndarray((2, 3)' in str(exc_info.value)

    def test_query(self, test_client: pxt.Client) -> None:
        cl = test_client
        col_names = ['c1', 'c2', 'c3', 'c4', 'c5']
        t = make_tbl(cl, 'test', col_names)
        rows = create_table_data(t)
        t.insert(rows)
        _ = t.show(n=0)

        # test querying existing table
        cl = pxt.Client(reload=True)
        t2 = cl.get_table('test')
        _  = t2.show(n=0)

    def test_update(self, test_tbl: pxt.Table, indexed_img_tbl: pxt.Table) -> None:
        t = test_tbl
        # update every type with a literal
        test_cases = [
            ('c1', 'new string'),
            # TODO: ('c1n', None),
            ('c3', -1.0),
            ('c4', True),
            ('c5', datetime.datetime.now()),
            ('c6', [{'x': 1, 'y': 2}]),
        ]
        count = t.count()
        for col_name, literal in test_cases:
            status = t.update({col_name: literal}, where=t.c3 < 10.0, cascade=False)
            assert status.num_rows == 10
            assert status.updated_cols == [f'{t.get_name()}.{col_name}']
            assert t.count() == count
            t.revert()

        # exchange two columns
        t.add_column(float_col=FloatType(nullable=True))
        t.update({'float_col': 1.0})
        float_col_vals = t.select(t.float_col).collect().to_pandas()['float_col']
        c3_vals = t.select(t.c3).collect().to_pandas()['c3']
        assert np.all(float_col_vals == pd.Series([1.0] * t.count()))
        t.update({'c3': t.float_col, 'float_col': t.c3})
        assert np.all(t.select(t.c3).collect().to_pandas()['c3'] == float_col_vals)
        assert np.all(t.select(t.float_col).collect().to_pandas()['float_col'] == c3_vals)
        t.revert()

        # update column that is used in computed cols
        t.add_column(computed1=t.c3 + 1)
        t.add_column(computed2=t.computed1 + 1)
        t.add_column(computed3=t.c3 + 3)

        # cascade=False
        computed1 = t.order_by(t.computed1).show(0).to_pandas()['computed1']
        computed2 = t.order_by(t.computed2).show(0).to_pandas()['computed2']
        computed3 = t.order_by(t.computed3).show(0).to_pandas()['computed3']
        assert t.where(t.c3 < 10.0).count() == 10
        assert t.where(t.c3 == 10.0).count() == 1
        # update to a value that also satisfies the where clause
        status = t.update({'c3': 0.0}, where=t.c3 < 10.0, cascade=False)
        assert status.num_rows == 10
        assert status.updated_cols == ['test_tbl.c3']
        assert t.where(t.c3 < 10.0).count() == 10
        assert t.where(t.c3 == 0.0).count() == 10
        # computed cols are not updated
        assert np.all(t.order_by(t.computed1).show(0).to_pandas()['computed1'] == computed1)
        assert np.all(t.order_by(t.computed2).show(0).to_pandas()['computed2'] == computed2)
        assert np.all(t.order_by(t.computed3).show(0).to_pandas()['computed3'] == computed3)

        # revert, then verify that we're back to where we started
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        t.revert()
        assert t.where(t.c3 < 10.0).count() == 10
        assert t.where(t.c3 == 10.0).count() == 1

        # cascade=True
        status = t.update({'c3': 0.0}, where=t.c3 < 10.0, cascade=True)
        assert status.num_rows == 10
        assert set(status.updated_cols) == \
               set(['test_tbl.c3', 'test_tbl.computed1', 'test_tbl.computed2', 'test_tbl.computed3'])
        assert t.where(t.c3 < 10.0).count() == 10
        assert t.where(t.c3 == 0.0).count() == 10
        assert np.all(t.order_by(t.computed1).show(0).to_pandas()['computed1'][:10] == pd.Series([1.0] * 10))
        assert np.all(t.order_by(t.computed2).show(0).to_pandas()['computed2'][:10] == pd.Series([2.0] * 10))
        assert np.all(t.order_by(t.computed3).show(0).to_pandas()['computed3'][:10] == pd.Series([3.0] * 10))

        # bad update spec
        with pytest.raises(excs.Error) as excinfo:
            t.update({1: 1})
        assert 'dict key' in str(excinfo.value)

        # unknown column
        with pytest.raises(excs.Error) as excinfo:
            t.update({'unknown': 1})
        assert 'unknown unknown' in str(excinfo.value)

        # incompatible type
        with pytest.raises(excs.Error) as excinfo:
            t.update({'c1': 1})
        assert 'not compatible' in str(excinfo.value)

        # can't update primary key
        with pytest.raises(excs.Error) as excinfo:
            t.update({'c2': 1})
        assert 'primary key' in str(excinfo.value)

        # can't update computed column
        with pytest.raises(excs.Error) as excinfo:
            t.update({'computed1': 1})
        assert 'is computed' in str(excinfo.value)

        # non-expr
        with pytest.raises(excs.Error) as excinfo:
            t.update({'c3': lambda c3: math.sqrt(c3)})
        assert 'not a recognized' in str(excinfo.value)

        # non-Predicate filter
        with pytest.raises(excs.Error) as excinfo:
            t.update({'c3': 1.0}, where=lambda c2: c2 == 10)
        assert 'Predicate' in str(excinfo.value)

        img_t = indexed_img_tbl

        # can't update image col
        with pytest.raises(excs.Error) as excinfo:
            img_t.update({'img': 17}, where=img_t.img.nearest('car'))
        assert 'has type image' in str(excinfo.value)

        # similarity search is not supported
        with pytest.raises(excs.Error) as excinfo:
            img_t.update({'split': 'train'}, where=img_t.img.nearest('car'))
        assert 'nearest()' in str(excinfo.value)

        # filter not expressible in SQL
        with pytest.raises(excs.Error) as excinfo:
            img_t.update({'split': 'train'}, where=img_t.img.width > 100)
        assert 'not expressible' in str(excinfo.value)

    def test_cascading_update(self, test_tbl: pxt.InsertableTable) -> None:
        t = test_tbl
        t.add_column(d1=t.c3 - 1)
        # add column that can be updated
        t.add_column(c10=FloatType(nullable=True))
        t.update({'c10': t.c3})
        # computed column that depends on two columns: exercise duplicate elimination during query construction
        t.add_column(d2=t.c3 - t.c10)
        r1 = t.where(t.c2 < 5).select(t.c3 + 1.0, t.c10 - 1.0, t.c3, 2.0).order_by(t.c2).show(0)
        t.update({'c4': True, 'c3': t.c3 + 1.0, 'c10': t.c10 - 1.0}, where=t.c2 < 5, cascade=True)
        r2 = t.where(t.c2 < 5).select(t.c3, t.c10, t.d1, t.d2).order_by(t.c2).show(0)
        assert_resultset_eq(r1, r2)

    def test_delete(self, test_tbl: pxt.Table, indexed_img_tbl: pxt.Table) -> None:
        t = test_tbl

        cnt = t.where(t.c3 < 10.0).count()
        assert cnt == 10
        cnt = t.where(t.c3 == 10.0).count()
        assert cnt == 1
        status = t.delete(where=t.c3 < 10.0)
        assert status.num_rows == 10
        cnt = t.where(t.c3 < 10.0).count()
        assert cnt == 0
        cnt = t.where(t.c3 == 10.0).count()
        assert cnt == 1

        # revert, then verify that we're back where we started
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        t.revert()
        cnt = t.where(t.c3 < 10.0).count()
        assert cnt == 10
        cnt = t.where(t.c3 == 10.0).count()
        assert cnt == 1

        # non-Predicate filter
        with pytest.raises(excs.Error) as excinfo:
            t.delete(where=lambda c2: c2 == 10)
        assert 'Predicate' in str(excinfo.value)

        img_t = indexed_img_tbl
        # similarity search is not supported
        with pytest.raises(excs.Error) as excinfo:
            img_t.delete(where=img_t.img.nearest('car'))
        assert 'nearest()' in str(excinfo.value)

        # filter not expressible in SQL
        with pytest.raises(excs.Error) as excinfo:
            img_t.delete(where=img_t.img.width > 100)
        assert 'not expressible' in str(excinfo.value)

    def test_computed_cols(self, test_client: pxt.client) -> None:
        cl = test_client
        schema = {
            'c1': IntType(nullable=False),
            'c2': FloatType(nullable=False),
            'c3': JsonType(nullable=False),
        }
        t : pxt.InsertableTable = cl.create_table('test', schema)
        status = t.add_column(c4=t.c1 + 1)
        assert status.num_excs == 0
        status = t.add_column(c5=t.c4 + 1)
        assert status.num_excs == 0
        status = t.add_column(c6=t.c1 / t.c2)
        assert status.num_excs == 0
        status = t.add_column(c7=t.c6 * t.c2)
        assert status.num_excs == 0
        status = t.add_column(c8=t.c3.detections['*'].bounding_box)
        assert status.num_excs == 0
        status = t.add_column(c9=lambda c2: math.sqrt(c2), type=FloatType())
        assert status.num_excs == 0

        # unstored cols that compute window functions aren't currently supported
        with pytest.raises((excs.Error)):
            t.add_column(c10=ptf.sum(t.c1, group_by=t.c1), stored=False)

        # Column.dependent_cols are computed correctly
        assert len(t.c1.col.dependent_cols) == 2
        assert len(t.c2.col.dependent_cols) == 3
        assert len(t.c3.col.dependent_cols) == 1
        assert len(t.c4.col.dependent_cols) == 1
        assert len(t.c5.col.dependent_cols) == 0
        assert len(t.c6.col.dependent_cols) == 1
        assert len(t.c7.col.dependent_cols) == 0
        assert len(t.c8.col.dependent_cols) == 0

        rows = create_table_data(t, ['c1', 'c2', 'c3'], num_rows=10)
        t.insert(rows)
        _ = t.show()

        # not allowed to pass values for computed cols
        with pytest.raises(excs.Error):
            rows2 = create_table_data(t, ['c1', 'c2', 'c3', 'c4'], num_rows=10)
            t.insert(rows2)

        # test loading from store
        cl = pxt.Client(reload=True)
        t = cl.get_table('test')
        assert len(t.columns()) == len(t.columns())
        for i in range(len(t.columns())):
            if t.columns()[i].value_expr is not None:
                assert t.columns()[i].value_expr.equals(t.columns()[i].value_expr)

        # make sure we can still insert data and that computed cols are still set correctly
        status = t.insert(rows)
        assert status.num_excs == 0
        res = t.show(0)
        tbl_df = t.show(0).to_pandas()

        # can't drop c4: c5 depends on it
        with pytest.raises(excs.Error):
            t.drop_column('c4')
        t.drop_column('c5')
        # now it works
        t.drop_column('c4')

    def test_expr_udf_computed_cols(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = cl.create_table('test', {'c1': IntType(nullable=False)})
        rows = [{'c1': i} for i in range(100)]
        status = t.insert(rows)
        assert status.num_rows == len(rows)
        status = t.add_column(c2=t.c1 + 1)
        assert status.num_excs == 0
        # call with positional arg
        status = t.add_column(c3=self.add1(t.c1))
        assert status.num_excs == 0
        # call with keyword arg
        status = t.add_column(c4=self.add1(a=t.c1))
        assert status.num_excs == 0

        # TODO: how to verify the output?
        describe_output = t.__repr__()
        # 'add1' didn't get swallowed/the expr udf is still visible in the column definition
        assert 'add1' in describe_output

        def check(t: pxt.Table) -> None:
            assert_resultset_eq(
                t.select(t.c1 + 1).order_by(t.c1).collect(),
                t.select(t.c2).order_by(t.c1).collect())
            assert_resultset_eq(
                t.select(t.c1 + 1).order_by(t.c1).collect(),
                t.select(t.c3).order_by(t.c1).collect())

        check(t)
        # test loading from store
        cl = pxt.Client(reload=True)
        t = cl.get_table('test')
        check(t)

        # make sure we can still insert data and that computed cols are still set correctly
        status = t.insert(rows)
        assert status.num_excs == 0
        check(t)

    def test_computed_col_exceptions(self, test_client: pxt.Client, test_tbl: catalog.Table) -> None:
        cl = test_client

        # exception during insert()
        schema = {'c2': IntType(nullable=False)}
        rows = list(test_tbl.select(test_tbl.c2).collect())
        t = cl.create_table('test_insert', schema)
        status = t.add_column(add1=self.f2(self.f1(t.c2)))
        assert status.num_excs == 0
        status = t.insert(rows, fail_on_exception=False)
        assert status.num_excs == 10
        assert 'test_insert.add1' in status.cols_with_excs
        assert t.where(t.add1.errortype != None).count() == 10

        # exception during add_column()
        t = cl.create_table('test_add_column', schema)
        status = t.insert(rows)
        assert status.num_rows == 100
        assert status.num_excs == 0
        status = t.add_column(add1=self.f2(self.f1(t.c2)))
        assert status.num_excs == 10
        assert 'test_add_column.add1' in status.cols_with_excs
        assert t.where(t.add1.errortype != None).count() == 10

    def _test_computed_img_cols(self, t: catalog.Table, stores_img_col: bool) -> None:
        rows = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
        rows = [{'img': r['img']} for r in rows[:20]]
        status = t.insert(rows)
        assert status.num_rows == 20
        _ = t.count()
        _ = t.show()
        assert MediaStore.count(t.get_id()) == t.count() * stores_img_col

        # test loading from store
        cl = pxt.Client(reload=True)
        t2 = cl.get_table(t.get_name())
        assert len(t.columns()) == len(t2.columns())
        for i in range(len(t.columns())):
            if t.columns()[i].value_expr is not None:
                assert t.columns()[i].value_expr.equals(t2.columns()[i].value_expr)

        # make sure we can still insert data and that computed cols are still set correctly
        t2.insert(rows)
        assert MediaStore.count(t2.get_id()) == t2.count() * stores_img_col
        res = t2.show(0)
        tbl_df = t2.show(0).to_pandas()

        # revert also removes computed images
        t2.revert()
        assert MediaStore.count(t2.get_id()) == t2.count() * stores_img_col

    def test_computed_img_cols(self, test_client: pxt.Client) -> None:
        cl = test_client
        schema = {'img': ImageType(nullable=False)}
        t = cl.create_table('test', schema)
        t.add_column(c2=t.img.width)
        # c3 is not stored by default
        t.add_column(c3=t.img.rotate(90))
        self._test_computed_img_cols(t, stores_img_col=False)

        t = cl.create_table('test2', schema)
        # c3 is now stored
        t.add_column(c3=t.img.rotate(90), stored=True)
        self._test_computed_img_cols(t, stores_img_col=True)
        _ = t[t.c3.errortype].show(0)

        # computed img col with exceptions
        t = cl.create_table('test3', schema)
        @pxt.udf(return_type=ImageType(), param_types=[ImageType()])
        def f(img: PIL.Image.Image) -> PIL.Image.Image:
            raise RuntimeError
        t.add_column(c3=f(t.img), stored=True)
        rows = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
        rows = [{'img': r['img']} for r in rows[:20]]
        t.insert(rows, fail_on_exception=False)
        _ = t[t.c3.errortype].show(0)

    def test_computed_window_fn(self, test_client: pxt.Client, test_tbl: catalog.Table) -> None:
        cl = test_client
        t = test_tbl
        # backfill
        t.add_column(c9=ptf.sum(t.c2, group_by=t.c4, order_by=t.c3))

        schema = {
            'c2': IntType(nullable=False),
            'c3': FloatType(nullable=False),
            'c4': BoolType(nullable=False),
        }
        new_t = cl.create_table('insert_test', schema)
        new_t.add_column(c5=lambda c2: c2 * c2, type=IntType())
        new_t.add_column(c6=ptf.sum(new_t.c5, group_by=new_t.c4, order_by=new_t.c3))
        rows = list(t.select(t.c2, t.c4, t.c3).collect())
        new_t.insert(rows)
        _ = new_t.show(0)

    def test_revert(self, test_client: pxt.Client) -> None:
        cl = test_client
        t1 = make_tbl(cl, 'test1', ['c1', 'c2'])
        assert t1.version() == 0
        rows1 = create_table_data(t1)
        t1.insert(rows1)
        assert t1.count() == len(rows1)
        assert t1.version() == 1
        rows2 = create_table_data(t1)
        t1.insert(rows2)
        assert t1.count() == len(rows1) + len(rows2)
        assert t1.version() == 2
        t1.revert()
        assert t1.count() == len(rows1)
        assert t1.version() == 1
        t1.insert(rows2)
        assert t1.count() == len(rows1) + len(rows2)
        assert t1.version() == 2

        # can't revert past version 0
        t1.revert()
        t1.revert()
        with pytest.raises(excs.Error) as excinfo:
            t1.revert()
        assert 'version 0' in str(excinfo.value)

    def test_add_column(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        num_orig_cols = len(t.columns())
        t.add_column(add1=pxt.IntType(nullable=True))
        assert len(t.columns()) == num_orig_cols + 1

        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(add2=pxt.IntType(nullable=False))
        assert 'cannot add non-nullable' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(add2=pxt.IntType(nullable=False), add3=pxt.StringType())
        assert 'requires exactly one keyword argument' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(pos=pxt.StringType(nullable=True))
        assert 'is reserved' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(add2=pxt.IntType(nullable=False), type=pxt.StringType())
        assert '"type" is redundant' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(add2=[[1.0, 2.0], [3.0, 4.0]], type=pxt.StringType())
        assert '"type" is redundant' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(add2=pxt.IntType(nullable=False), stored=False)
        assert 'stored=false only applies' in str(exc_info.value).lower()

        # duplicate name
        with pytest.raises(excs.Error) as exc_info:
            _ = t.add_column(c1=pxt.IntType())
        assert 'duplicate column name' in str(exc_info.value).lower()

        # 'stored' kwarg only applies to computed image columns
        with pytest.raises(excs.Error):
            _ = t.add_column(c5=IntType(), stored=False)
        with pytest.raises(excs.Error):
            _ = t.add_column(c5=ImageType(), stored=False)
        with pytest.raises(excs.Error):
            _ = t.add_column(c5=(t.c2 + t.c3), stored=False)

        # make sure this is still true after reloading the metadata
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        assert len(t.columns()) == num_orig_cols + 1

        # revert() works
        t.revert()
        assert len(t.columns()) == num_orig_cols

        # make sure this is still true after reloading the metadata once more
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        assert len(t.columns()) == num_orig_cols

    def test_add_column_setitem(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        num_orig_cols = len(t.columns())
        t['add1'] = pxt.IntType(nullable=True)
        assert len(t.columns()) == num_orig_cols + 1
        t['computed1'] = t.c2 + 1
        assert len(t.columns()) == num_orig_cols + 2

        with pytest.raises(excs.Error) as exc_info:
            _ = t['pos'] = pxt.StringType()
        assert 'is reserved' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t[2] = pxt.StringType()
        assert 'must be a string' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t['add 2'] = pxt.StringType()
        assert 'invalid column name' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t['add2'] = {'value': t.c2 + 1, 'type': pxt.StringType()}
        assert '"type" is redundant' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t['add2'] = {'value': pxt.IntType()}
        assert 'value needs to be either' in str(exc_info.value).lower()

        with pytest.raises(excs.Error) as exc_info:
            _ = t['add2'] = {'value': t.c2 + 1, 'stored': False}
        assert 'stored=false only applies' in str(exc_info.value).lower()

        # duplicate name
        with pytest.raises(excs.Error) as exc_info:
            _ = t['c1'] = pxt.IntType()
        assert 'duplicate column name' in str(exc_info.value).lower()

        # make sure this is still true after reloading the metadata
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        assert len(t.columns()) == num_orig_cols + 2

        # revert() works
        t.revert()
        t.revert()
        assert len(t.columns()) == num_orig_cols

        # make sure this is still true after reloading the metadata once more
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        assert len(t.columns()) == num_orig_cols

    def test_drop_column(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        num_orig_cols = len(t.columns())
        t.drop_column('c1')
        assert len(t.columns()) == num_orig_cols - 1

        with pytest.raises(excs.Error):
            t.drop_column('unknown')

        # make sure this is still true after reloading the metadata
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        assert len(t.columns()) == num_orig_cols - 1

        # revert() works
        t.revert()
        assert len(t.columns()) == num_orig_cols

        # make sure this is still true after reloading the metadata once more
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        assert len(t.columns()) == num_orig_cols

    def test_rename_column(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        num_orig_cols = len(t.columns())
        t.rename_column('c1', 'c1_renamed')
        assert len(t.columns()) == num_orig_cols

        def check_rename(t: pxt.Table, known: str, unknown: str) -> None:
            with pytest.raises(AttributeError) as exc_info:
                _ = t.select(t[unknown]).collect()
            assert 'unknown' in str(exc_info.value).lower()
            _ = t.select(t[known]).collect()

        check_rename(t, 'c1_renamed', 'c1')

        # unknown column
        with pytest.raises(excs.Error):
            t.rename_column('unknown', 'unknown_renamed')
        # bad name
        with pytest.raises(excs.Error):
            t.rename_column('c2', 'bad name')
        # existing name
        with pytest.raises(excs.Error):
            t.rename_column('c2', 'c3')

        # make sure this is still true after reloading the metadata
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        check_rename(t, 'c1_renamed', 'c1')

        # revert() works
        t.revert()
        _ = t.select(t.c1).collect()
        #check_rename(t, 'c1', 'c1_renamed')

        # make sure this is still true after reloading the metadata once more
        cl = pxt.Client(reload=True)
        t = cl.get_table(t.get_name())
        check_rename(t, 'c1', 'c1_renamed')

    def test_add_computed_column(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        status = t.add_column(add1=t.c2 + 10)
        assert status.num_excs == 0
        _ = t.show()

        # with exception in SQL
        with pytest.raises(excs.Error):
            t.add_column(add2=(t.c2 - 10) / (t.c3 - 10))

        # with exception in Python for c6.f2 == 10
        status = t.add_column(add2=(t.c6.f2 - 10) / (t.c6.f2 - 10))
        assert status.num_excs == 1
        result = t[t.add2.errortype != None][t.c6.f2, t.add2, t.add2.errortype, t.add2.errormsg].show()
        assert len(result) == 1

        # test case: exceptions in dependencies prevent execution of dependent exprs
        status = t.add_column(add3=self.f2(self.f1(t.c2)))
        assert status.num_excs == 10
        result = t[t.add3.errortype != None][t.c2, t.add3, t.add3.errortype, t.add3.errormsg].show()
        assert len(result) == 10

    def test_describe(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        fn = lambda c2: np.full((3, 4), c2)
        t.add_column(computed1=fn, type=ArrayType((3, 4), dtype=IntType()))
        t.describe()
        t.comment = 'This is a comment.'
        t.describe()

        # TODO: how to you check the output of these?
        _ = repr(t)
        _ = t._repr_html_()

    def test_common_col_names(self, test_client: pxt.Client) -> None:
        """Make sure that commonly used column names don't collide with Table member vars"""
        cl = test_client
        schema = {'id': IntType(nullable=False), 'name': StringType(nullable=False)}
        tbl = cl.create_table('test', schema)
        status = tbl.insert({'id': id, 'name': str(id)} for id in range(10))
        assert status.num_rows == 10
        assert status.num_excs == 0
        assert tbl.count() == 10
        # we can create references to those column via __getattr__
        _ = tbl.select(tbl.id, tbl.name).collect()



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple

import sqlalchemy as sql

from .globals import ArithmeticOperator
from .expr import Expr
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.exceptions as excs
import pixeltable.catalog as catalog
import pixeltable.type_system as ts


class ArithmeticExpr(Expr):
    """
    Allows arithmetic exprs on json paths
    """
    def __init__(self, operator: ArithmeticOperator, op1: Expr, op2: Expr):
        # TODO: determine most specific common supertype
        if op1.col_type.is_json_type() or op2.col_type.is_json_type():
            # we assume it's a float
            super().__init__(ts.FloatType())
        else:
            super().__init__(ts.ColumnType.supertype(op1.col_type, op2.col_type))
        self.operator = operator
        self.components = [op1, op2]

        # do typechecking after initialization in order for __str__() to work
        if not op1.col_type.is_numeric_type() and not op1.col_type.is_json_type():
            raise excs.Error(f'{self}: {operator} requires numeric types, but {op1} has type {op1.col_type}')
        if not op2.col_type.is_numeric_type() and not op2.col_type.is_json_type():
            raise excs.Error(f'{self}: {operator} requires numeric types, but {op2} has type {op2.col_type}')

        self.id = self._create_id()

    def __str__(self) -> str:
        # add parentheses around operands that are ArithmeticExprs to express precedence
        op1_str = f'({self._op1})' if isinstance(self._op1, ArithmeticExpr) else str(self._op1)
        op2_str = f'({self._op2})' if isinstance(self._op2, ArithmeticExpr) else str(self._op2)
        return f'{op1_str} {str(self.operator)} {op2_str}'

    def _equals(self, other: ArithmeticExpr) -> bool:
        return self.operator == other.operator

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('operator', self.operator.value)]

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        left = self._op1.sql_expr()
        right = self._op2.sql_expr()
        if left is None or right is None:
            return None
        if self.operator == ArithmeticOperator.ADD:
            return left + right
        if self.operator == ArithmeticOperator.SUB:
            return left - right
        if self.operator == ArithmeticOperator.MUL:
            return left * right
        if self.operator == ArithmeticOperator.DIV:
            return left / right
        if self.operator == ArithmeticOperator.MOD:
            return left % right

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        op1_val = data_row[self._op1.slot_idx]
        op2_val = data_row[self._op2.slot_idx]
        # check types if we couldn't do that prior to execution
        if self._op1.col_type.is_json_type() and not isinstance(op1_val, int) and not isinstance(op1_val, float):
            raise excs.Error(
                f'{self.operator} requires numeric type, but {self._op1} has type {type(op1_val).__name__}')
        if self._op2.col_type.is_json_type() and not isinstance(op2_val, int) and not isinstance(op2_val, float):
            raise excs.Error(
                f'{self.operator} requires numeric type, but {self._op2} has type {type(op2_val).__name__}')

        if self.operator == ArithmeticOperator.ADD:
            data_row[self.slot_idx] = op1_val + op2_val
        elif self.operator == ArithmeticOperator.SUB:
            data_row[self.slot_idx] = op1_val - op2_val
        elif self.operator == ArithmeticOperator.MUL:
            data_row[self.slot_idx] = op1_val * op2_val
        elif self.operator == ArithmeticOperator.DIV:
            data_row[self.slot_idx] = op1_val / op2_val
        elif self.operator == ArithmeticOperator.MOD:
            data_row[self.slot_idx] = op1_val % op2_val

    def _as_dict(self) -> Dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'operator' in d
        assert len(components) == 2
        return cls(ArithmeticOperator(d['operator']), components[0], components[1])



from typing import List, Optional, Dict, Type, Any, Union
import pandas as pd
import logging
import dataclasses
from uuid import UUID
from collections import defaultdict

import sqlalchemy as sql
import sqlalchemy.orm as orm

from pixeltable.metadata import schema
from pixeltable.env import Env
import pixeltable.func as func
import pixeltable.catalog as catalog
from pixeltable import exceptions as excs
from pixeltable.exprs import Predicate
from pixeltable.iterators import ComponentIterator

__all__ = [
    'Client',
]


_logger = logging.getLogger('pixeltable')

class Client:
    """
    Client for interacting with a Pixeltable environment.
    """

    def __init__(self, reload: bool = False) -> None:
        """Constructs a client.
        """
        env = Env.get()
        env.set_up()
        env.upgrade_metadata()
        if reload:
            catalog.Catalog.clear()
        self.catalog = catalog.Catalog.get()

    def logging(
            self, *, to_stdout: Optional[bool] = None, level: Optional[int] = None,
            add: Optional[str] = None, remove: Optional[str] = None
    ) -> None:
        """Configure logging.

        Args:
            to_stdout: if True, also log to stdout
            level: default log level
            add: comma-separated list of 'module name:log level' pairs; ex.: add='video:10'
            remove: comma-separated list of module names
        """
        if to_stdout is not None:
            Env.get().log_to_stdout(to_stdout)
        if level is not None:
            Env.get().set_log_level(level)
        if add is not None:
            for module, level in [t.split(':') for t in add.split(',')]:
                Env.get().set_module_log_level(module, int(level))
        if remove is not None:
            for module in remove.split(','):
                Env.get().set_module_log_level(module, None)
        if to_stdout is None and level is None and add is None and remove is None:
            Env.get().print_log_config()

    def list_functions(self) -> pd.DataFrame:
        """Returns information about all registered functions.

        Returns:
            Pandas DataFrame with columns 'Path', 'Name', 'Parameters', 'Return Type', 'Is Agg', 'Library'
        """
        functions = func.FunctionRegistry.get().list_functions()
        paths = ['.'.join(f.self_path.split('.')[:-1]) for f in functions]
        names = [f.name for f in functions]
        params = [
            ', '.join(
                [param_name + ': ' + str(param_type) for param_name, param_type in f.signature.parameters.items()])
            for f in functions
        ]
        pd_df = pd.DataFrame({
            'Path': paths,
            'Function Name': names,
            'Parameters': params,
            'Return Type': [str(f.signature.get_return_type()) for f in functions],
        })
        pd_df = pd_df.style.set_properties(**{'text-align': 'left'}) \
            .set_table_styles([dict(selector='th', props=[('text-align', 'center')])])  # center-align headings
        return pd_df.hide(axis='index')

    def get_path(self, schema_obj: catalog.SchemaObject) -> str:
        """Returns the path to a SchemaObject.

        Args:
            schema_obj: SchemaObject to get the path for.

        Returns:
            Path to the SchemaObject.
        """
        path_elements: List[str] = []
        dir_id = schema_obj._dir_id
        while dir_id is not None:
            dir = self.catalog.paths.get_schema_obj(dir_id)
            if dir._dir_id is None:
                # this is the root dir with name '', which we don't want to include in the path
                break
            path_elements.insert(0, dir._name)
            dir_id = dir._dir_id
        path_elements.append(schema_obj._name)
        return '.'.join(path_elements)

    def create_table(
            self, path_str: str, schema: Dict[str, Any], primary_key: Optional[Union[str, List[str]]] = None,
            num_retained_versions: int = 10, comment: str = ''
    ) -> catalog.InsertableTable:
        """Create a new `InsertableTable`.

        Args:
            path_str: Path to the table.
            schema: dictionary mapping column names to column types, value expressions, or to column specifications.
            num_retained_versions: Number of versions of the table to retain.

        Returns:
            The newly created table.

        Raises:
            Error: if the path already exists or is invalid.

        Examples:
            Create a table with an int and a string column:

            >>> table = cl.create_table('my_table', schema={'col1': IntType(), 'col2': StringType()})

            Create a table with a single indexed image column:

            >>> table = cl.create_table('my_table', schema={'col1': {'type': ImageType(), 'indexed': True}})
        """
        path = catalog.Path(path_str)
        self.catalog.paths.check_is_valid(path, expected=None)
        dir = self.catalog.paths[path.parent]

        if len(schema) == 0:
            raise excs.Error(f'Table schema is empty: `{path_str}`')

        if primary_key is None:
            primary_key = []
        elif isinstance(primary_key, str):
            primary_key = [primary_key]
        else:
            if not isinstance(primary_key, list) or not all(isinstance(pk, str) for pk in primary_key):
                raise excs.Error('primary_key must be a single column name or a list of column names')

        tbl = catalog.InsertableTable.create(
            dir._id, path.name, schema, primary_key=primary_key, num_retained_versions=num_retained_versions, comment=comment)
        self.catalog.paths[path] = tbl
        _logger.info(f'Created table `{path_str}`.')
        return tbl

    def create_view(
            self, path_str: str, base: catalog.Table, *, schema: Optional[Dict[str, Any]] = None,
            filter: Optional[Predicate] = None,
            is_snapshot: bool = False, iterator_class: Optional[Type[ComponentIterator]] = None,
            iterator_args: Optional[Dict[str, Any]] = None, num_retained_versions: int = 10, comment: str = '',
            ignore_errors: bool = False) -> catalog.View:
        """Create a new `View`.

        Args:
            path_str: Path to the view.
            base: Table (ie, table or view or snapshot) to base the view on.
            schema: dictionary mapping column names to column types, value expressions, or to column specifications.
            filter: Predicate to filter rows of the base table.
            is_snapshot: Whether the view is a snapshot.
            iterator_class: Class of the iterator to use for the view.
            iterator_args: Arguments to pass to the iterator class.
            num_retained_versions: Number of versions of the view to retain.
            ignore_errors: if True, fail silently if the path already exists or is invalid.

        Returns:
            The newly created view.

        Raises:
            Error: if the path already exists or is invalid.

        Examples:
            Create a view with an additional int and a string column and a filter:

            >>> view = cl.create_view(
                'my_view', base, schema={'col3': IntType(), 'col4': StringType()}, filter=base.col1 > 10)

            Create a table snapshot:

            >>> snapshot_view = cl.create_view('my_snapshot_view', base, is_snapshot=True)

            Create an immutable view with additional computed columns and a filter:

            >>> snapshot_view = cl.create_view(
                'my_snapshot', base, schema={'col3': base.col2 + 1}, filter=base.col1 > 10, is_snapshot=True)
        """
        assert (iterator_class is None) == (iterator_args is None)
        assert isinstance(base, catalog.Table)
        path = catalog.Path(path_str)
        try:
            self.catalog.paths.check_is_valid(path, expected=None)
        except Exception as e:
            if ignore_errors:
                return
            else:
                raise e
        dir = self.catalog.paths[path.parent]

        if schema is None:
            schema = {}
        view = catalog.View.create(
            dir._id, path.name, base=base, schema=schema, predicate=filter, is_snapshot=is_snapshot,
            iterator_cls=iterator_class, iterator_args=iterator_args, num_retained_versions=num_retained_versions, comment=comment)
        self.catalog.paths[path] = view
        _logger.info(f'Created view `{path_str}`.')
        return view

    def get_table(self, path: str) -> catalog.Table:
        """Get a handle to a table (including views and snapshots).

        Args:
            path: Path to the table.

        Returns:
            A `InsertableTable` or `View` object.

        Raises:
            Error: If the path does not exist or does not designate a table.

        Examples:
            Get handle for a table in the top-level directory:

            >>> table = cl.get_table('my_table')

            For a table in a subdirectory:

            >>> table = cl.get_table('subdir.my_table')

            For a snapshot in the top-level directory:

            >>> table = cl.get_table('my_snapshot')
        """
        p = catalog.Path(path)
        self.catalog.paths.check_is_valid(p, expected=catalog.Table)
        obj = self.catalog.paths[p]
        return obj

    def move(self, path: str, new_path: str) -> None:
        """Move a schema object to a new directory and/or rename a schema object.

        Args:
            path: absolute path to the existing schema object.
            new_path: absolute new path for the schema object.

        Raises:
            Error: If path does not exist or new_path already exists.

        Examples:
            Move a table to a different directory:

            >>>> cl.move('dir1.my_table', 'dir2.my_table')

            Rename a table:

            >>>> cl.move('dir1.my_table', 'dir1.new_name')
        """
        p = catalog.Path(path)
        self.catalog.paths.check_is_valid(p, expected=catalog.SchemaObject)
        new_p = catalog.Path(new_path)
        self.catalog.paths.check_is_valid(new_p, expected=None)
        obj = self.catalog.paths[p]
        self.catalog.paths.move(p, new_p)
        new_dir = self.catalog.paths[new_p.parent]
        obj.move(new_p.name, new_dir._id)

    def list_tables(self, dir_path: str = '', recursive: bool = True) -> List[str]:
        """List the tables in a directory.

        Args:
            dir_path: Path to the directory. Defaults to the root directory.
            recursive: Whether to list tables in subdirectories as well.

        Returns:
            A list of table paths.

        Raises:
            Error: If the path does not exist or does not designate a directory.

        Examples:
            List tables in top-level directory:

            >>> cl.list_tables()
            ['my_table', ...]

            List tables in 'dir1':

            >>> cl.list_tables('dir1')
            [...]
        """
        assert dir_path is not None
        path = catalog.Path(dir_path, empty_is_valid=True)
        self.catalog.paths.check_is_valid(path, expected=catalog.Dir)
        return [str(p) for p in self.catalog.paths.get_children(path, child_type=catalog.Table, recursive=recursive)]

    def drop_table(self, path: str, force: bool = False, ignore_errors: bool = False) -> None:
        """Drop a table.

        Args:
            path: Path to the table.
            force: Whether to drop the table even if it has unsaved changes.
            ignore_errors: Whether to ignore errors if the table does not exist.

        Raises:
            Error: If the path does not exist or does not designate a table and ignore_errors is False.

        Examples:
            >>> cl.drop_table('my_table')
        """
        path_obj = catalog.Path(path)
        try:
            self.catalog.paths.check_is_valid(path_obj, expected=catalog.Table)
        except Exception as e:
            if ignore_errors:
                _logger.info(f'Skipped table `{path}` (does not exist).')
                return
            else:
                raise e
        tbl = self.catalog.paths[path_obj]
        if len(self.catalog.tbl_dependents[tbl._id]) > 0:
            dependent_paths = [self.get_path(dep) for dep in self.catalog.tbl_dependents[tbl._id]]
            raise excs.Error(f'Table {path} has dependents: {", ".join(dependent_paths)}')
        tbl._drop()
        del self.catalog.paths[path_obj]
        _logger.info(f'Dropped table `{path}`.')

    def create_dir(self, path_str: str, ignore_errors: bool = False) -> None:
        """Create a directory.

        Args:
            path_str: Path to the directory.
            ignore_errors: if True, silently returns on error

        Raises:
            Error: If the path already exists or the parent is not a directory.

        Examples:
            >>> cl.create_dir('my_dir')

            Create a subdirectory:

            >>> cl.create_dir('my_dir.sub_dir')
        """
        try:
            path = catalog.Path(path_str)
            self.catalog.paths.check_is_valid(path, expected=None)
            parent = self.catalog.paths[path.parent]
            assert parent is not None
            with orm.Session(Env.get().engine, future=True) as session:
                dir_md = schema.DirMd(name=path.name)
                dir_record = schema.Dir(parent_id=parent._id, md=dataclasses.asdict(dir_md))
                session.add(dir_record)
                session.flush()
                assert dir_record.id is not None
                self.catalog.paths[path] = catalog.Dir(dir_record.id, parent._id, path.name)
                session.commit()
                _logger.info(f'Created directory `{path_str}`.')
                print(f'Created directory `{path_str}`.')
        except excs.Error as e:
            if ignore_errors:
                return
            else:
                raise e

    def rm_dir(self, path_str: str) -> None:
        """Remove a directory.

        Args:
            path_str: Path to the directory.

        Raises:
            Error: If the path does not exist or does not designate a directory or if the directory is not empty.

        Examples:
            >>> cl.rm_dir('my_dir')

            Remove a subdirectory:

            >>> cl.rm_dir('my_dir.sub_dir')
        """
        path = catalog.Path(path_str)
        self.catalog.paths.check_is_valid(path, expected=catalog.Dir)

        # make sure it's empty
        if len(self.catalog.paths.get_children(path, child_type=None, recursive=True)) > 0:
            raise excs.Error(f'Directory {path_str} is not empty')
        # TODO: figure out how to make force=True work in the presence of snapshots
        #        # delete tables
        #        for tbl_path in self.paths.get_children(path, child_type=MutableTable, recursive=True):
        #            self.drop_table(str(tbl_path), force=True)
        #        # rm subdirs
        #        for dir_path in self.paths.get_children(path, child_type=Dir, recursive=False):
        #            self.rm_dir(str(dir_path), force=True)

        with Env.get().engine.begin() as conn:
            dir = self.catalog.paths[path]
            conn.execute(sql.delete(schema.Dir.__table__).where(schema.Dir.id == dir._id))
        del self.catalog.paths[path]
        _logger.info(f'Removed directory {path_str}')

    def list_dirs(self, path_str: str = '', recursive: bool = True) -> List[str]:
        """List the directories in a directory.

        Args:
            path_str: Path to the directory.
            recursive: Whether to list subdirectories recursively.

        Returns:
            List of directory paths.

        Raises:
            Error: If the path does not exist or does not designate a directory.

        Examples:
            >>> cl.list_dirs('my_dir', recursive=True)
            ['my_dir', 'my_dir.sub_dir1']
        """
        path = catalog.Path(path_str, empty_is_valid=True)
        self.catalog.paths.check_is_valid(path, expected=catalog.Dir)
        return [str(p) for p in self.catalog.paths.get_children(path, child_type=catalog.Dir, recursive=recursive)]

    # TODO: for now, named functions are deprecated, until we understand the use case and requirements better
    # def create_function(self, path_str: str, fn: func.Function) -> None:
    #     """Create a stored function.
    #
    #     Args:
    #         path_str: path where the function gets stored
    #         func: previously created Function object
    #
    #     Raises:
    #         Error: if the path already exists or the parent is not a directory
    #
    #     Examples:
    #         Create a function ``detect()`` that takes an image and returns a JSON object, and store it in ``my_dir``:
    #
    #         >>> @pxt.udf(param_types=[ImageType()], return_type=JsonType())
    #         ... def detect(img):
    #         ... ...
    #         >>> cl.create_function('my_dir.detect', detect)
    #     """
    #     if fn.is_module_function:
    #         raise excs.Error(f'Cannot create a named function for a library function')
    #     path = catalog.Path(path_str)
    #     self.catalog.paths.check_is_valid(path, expected=None)
    #     dir = self.catalog.paths[path.parent]
    #
    #     func.FunctionRegistry.get().create_function(fn, dir._id, path.name)
    #     self.catalog.paths[path] = catalog.NamedFunction(fn.id, dir._id, path.name)
    #     fn.md.fqn = str(path)
    #     _logger.info(f'Created function {path_str}')
    #
    # def update_function(self, path_str: str, fn: func.Function) -> None:
    #     """Update the implementation of a stored function.
    #
    #     Args:
    #         path_str: path to the function to be updated
    #         func: new function implementation
    #
    #     Raises:
    #         Error: if the path does not exist or ``func`` has a different signature than the stored function.
    #     """
    #     if fn.is_module_function:
    #         raise excs.Error(f'Cannot update a named function to a library function')
    #     path = catalog.Path(path_str)
    #     self.catalog.paths.check_is_valid(path, expected=catalog.NamedFunction)
    #     named_fn = self.catalog.paths[path]
    #     f = func.FunctionRegistry.get().get_function(id=named_fn._id)
    #     if f.md.signature != fn.md.signature:
    #         raise excs.Error(
    #             f'The function signature cannot be changed. The existing signature is {f.md.signature}')
    #     if f.is_aggregate != fn.is_aggregate:
    #         raise excs.Error(f'Cannot change an aggregate function into a non-aggregate function and vice versa')
    #     func.FunctionRegistry.get().update_function(named_fn._id, fn)
    #     _logger.info(f'Updated function {path_str}')
    #
    # def get_function(self, path_str: str) -> func.Function:
    #     """Get a handle to a stored function.
    #
    #     Args:
    #         path_str: path to the function
    #
    #     Returns:
    #         Function object
    #
    #     Raises:
    #         Error: if the path does not exist or is not a function
    #
    #     Examples:
    #         >>> detect = cl.get_function('my_dir.detect')
    #     """
    #     path = catalog.Path(path_str)
    #     self.catalog.paths.check_is_valid(path, expected=catalog.NamedFunction)
    #     named_fn = self.catalog.paths[path]
    #     assert isinstance(named_fn, catalog.NamedFunction)
    #     fn = func.FunctionRegistry.get().get_function(id=named_fn._id)
    #     fn.md.fqn = str(path)
    #     return fn
    #
    # def drop_function(self, path_str: str, ignore_errors: bool = False) -> None:
    #     """Deletes stored function.
    #
    #     Args:
    #         path_str: path to the function
    #         ignore_errors: if True, does not raise if the function does not exist
    #
    #     Raises:
    #         Error: if the path does not exist or is not a function
    #
    #     Examples:
    #         >>> cl.drop_function('my_dir.detect')
    #     """
    #     path = catalog.Path(path_str)
    #     try:
    #         self.catalog.paths.check_is_valid(path, expected=catalog.NamedFunction)
    #     except excs.Error as e:
    #         if ignore_errors:
    #             return
    #         else:
    #             raise e
    #     named_fn = self.catalog.paths[path]
    #     func.FunctionRegistry.get().delete_function(named_fn._id)
    #     del self.catalog.paths[path]
    #     _logger.info(f'Dropped function {path_str}')




import datetime
import pickle
from pathlib import Path
from typing import Any, Dict

import bs4
import numpy as np
import pytest
import requests
from pycocotools.coco import COCO

import pixeltable as pxt
from pixeltable import catalog
from pixeltable import exceptions as excs
from pixeltable.iterators import FrameIterator
from pixeltable.tests.utils import get_video_files, get_audio_files, skip_test_if_not_installed


class TestDataFrame:
    def test_select_where(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        res1 = t[t.c1, t.c2, t.c3].show(0)
        res2 = t.select(t.c1, t.c2, t.c3).show(0)
        assert res1 == res2

        res1 = t[t.c2 < 10][t.c1, t.c2, t.c3].show(0)
        res2 = t.where(t.c2 < 10).select(t.c1, t.c2, t.c3).show(0)
        assert res1 == res2

        res3 = t.where(t.c2 < 10).select(c1=t.c1, c2=t.c2, c3=t.c3).show(0)
        assert res1 == res3

        res4 = t.where(t.c2 < 10).select(t.c1, c2=t.c2, c3=t.c3).show(0)
        assert res1 == res4

        _ = t.where(t.c2 < 10).select(t.c2, t.c2).show(0) # repeated name no error
        
        # duplicate select list
        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c1).select(t.c2).show(0)
        assert 'already specified' in str(exc_info.value)

        # invalid expr in select list: Callable is not a valid literal
        with pytest.raises(TypeError) as exc_info:
            _ = t.select(datetime.datetime.now).show(0)
        assert 'Not a valid literal' in str(exc_info.value)

        # catch invalid name in select list from user input
        # only check stuff that's not caught by python kwargs checker
        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c1, **{'c2-1': t.c2}).show(0)
        assert 'Invalid name' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c1, **{'': t.c2}).show(0)
        assert 'Invalid name' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c1, **{'foo.bar': t.c2}).show(0)
        assert 'Invalid name' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c1, _c3=t.c2).show(0)
        assert 'Invalid name' in str(exc_info.value)

        # catch repeated name from user input
        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c2, c2=t.c1).show(0)
        assert 'Repeated column name' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = t.select(t.c2+1, col_0=t.c2).show(0)
        assert 'Repeated column name' in str(exc_info.value)

    def test_result_set_iterator(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        res = t.select(t.c1, t.c2, t.c3).collect()
        pd_df = res.to_pandas()

        def check_row(row: Dict[str, Any], idx: int) -> None:
            assert len(row) == 3
            assert 'c1' in row
            assert row['c1'] == pd_df['c1'][idx]
            assert 'c2' in row
            assert row['c2'] == pd_df['c2'][idx]
            assert 'c3' in row
            assert row['c3'] == pd_df['c3'][idx]

        # row iteration
        for idx, row in enumerate(res):
            check_row(row, idx)

        # row access
        row = res[0]
        check_row(row, 0)

        # column access
        col_values = res['c2']
        assert col_values == pd_df['c2'].values.tolist()

        # cell access
        assert res[0, 'c2'] == pd_df['c2'][0]
        assert res[0, 'c2'] == res[0, 1]

        with pytest.raises(excs.Error) as exc_info:
            _ = res['does_not_exist']
        assert 'Invalid column name' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = res[0, 'does_not_exist']
        assert 'Invalid column name' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = res[0, 0, 0]
        assert 'Bad index' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = res['c2', 0]
        assert 'Bad index' in str(exc_info.value)

    def test_order_by(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        res = t.select(t.c4, t.c2).order_by(t.c4).order_by(t.c2, asc=False).show(0)

        # invalid expr in order_by()
        with pytest.raises(excs.Error) as exc_info:
            _ = t.order_by(datetime.datetime.now()).show(0)
        assert 'Invalid expression' in str(exc_info.value)

    def test_head_tail(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        res = t.head(10).to_pandas()
        assert np.all(res.c2 == list(range(10)))
        # Where is applied
        res = t.where(t.c2 > 9).head(10).to_pandas()
        assert np.all(res.c2 == list(range(10, 20)))
        # order_by() is an error
        with pytest.raises(excs.Error) as exc_info:
            _ = t.order_by(t.c2).head(10)
        assert 'cannot be used with order_by' in str(exc_info.value)

        res = t.tail().to_pandas()
        assert np.all(res.c2 == list(range(90, 100)))
        res = t.where(t.c2 < 90).tail().to_pandas()
        assert np.all(res.c2 == list(range(80, 90)))
        # order_by() is an error
        with pytest.raises(excs.Error) as exc_info:
            _ = t.order_by(t.c2).tail(10)
        assert 'cannot be used with order_by' in str(exc_info.value)

    def test_describe(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        df = t.select(t.c1).where(t.c2 < 10).limit(10)
        df.describe()

        # TODO: how to you check the output of these?
        _ = df.__repr__()
        _ = df._repr_html_()

    def test_count(self, test_tbl: catalog.Table, indexed_img_tbl: catalog.Table) -> None:
        skip_test_if_not_installed('nos')
        t = test_tbl
        cnt = t.count()
        assert cnt == 100

        cnt = t.where(t.c2 < 10).count()
        assert cnt == 10

        # count() doesn't work with similarity search
        t = indexed_img_tbl
        probe = t.select(t.img).show(1)
        img = probe[0, 0]
        with pytest.raises(excs.Error):
            _ = t.where(t.img.nearest(img)).count()
        with pytest.raises(excs.Error):
            _ = t.where(t.img.nearest('car')).count()

        # for now, count() doesn't work with non-SQL Where clauses
        with pytest.raises(excs.Error):
            _ = t.where(t.img.width > 100).count()

    def test_select_literal(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        res = t.select(1.0).where(t.c2 < 10).collect()
        assert res[res.column_names()[0]] == [1.0] * 10

    def test_html_media_url(self, test_client: pxt.Client) -> None:
        tab = test_client.create_table('test_html_repr', {'video': pxt.VideoType(), 'audio': pxt.AudioType()})
        status = tab.insert(video=get_video_files()[0], audio=get_audio_files()[0])
        assert status.num_rows == 1
        assert status.num_excs == 0

        res = tab.select(tab.video, tab.audio).collect()
        doc = bs4.BeautifulSoup(res._repr_html_(), features='html.parser')
        video_tags = doc.find_all('video')
        assert len(video_tags) == 1
        audio_tags = doc.find_all('audio')
        assert len(audio_tags) == 1

        # get the source elements and test their src attributes
        for tag in video_tags + audio_tags:
            sources = tag.find_all('source')
            assert len(sources) == 1
            for src in sources:
                response = requests.get(src['src'])
                assert response.status_code == 200

    def test_to_pytorch_dataset(self, all_datatypes_tbl: catalog.Table):
        """ tests all types are handled correctly in this conversion
        """
        import torch

        t = all_datatypes_tbl
        df = t.where(t.row_id < 1)
        assert df.count() > 0
        ds = df.to_pytorch_dataset()
        type_dict = dict(zip(df.get_column_names(),df.get_column_types()))
        for tup in ds:
            for col in df.get_column_names():
                assert col in tup
        
            arrval = tup['c_array']
            assert isinstance(arrval, np.ndarray)
            col_type = type_dict['c_array']
            assert arrval.dtype == col_type.numpy_dtype()
            assert arrval.shape == col_type.shape
            assert arrval.dtype == np.float32
            assert arrval.flags["WRITEABLE"], 'required by pytorch collate function'

            assert isinstance(tup['c_bool'], bool)
            assert isinstance(tup['c_int'], int)
            assert isinstance(tup['c_float'], float)
            assert isinstance(tup['c_timestamp'], float)
            assert torch.is_tensor(tup['c_image'])
            assert isinstance(tup['c_video'], str)
            assert isinstance(tup['c_json'], dict)

    def test_to_pytorch_image_format(self, all_datatypes_tbl: catalog.Table) -> None:
        """ tests the image_format parameter is honored
        """
        import torch
        import torchvision.transforms as T

        W, H = 220, 224 # make different from each other
        t = all_datatypes_tbl
        df = t.select(
            t.row_id,
            t.c_image,
            c_image_xformed=t.c_image.resize([W, H]).convert('RGB')
        ).where(t.row_id < 1)

        pandas_df = df.show().to_pandas()
        im_plain = pandas_df['c_image'].values[0]
        im_xformed = pandas_df['c_image_xformed'].values[0]
        assert pandas_df.shape[0] == 1

        ds = df.to_pytorch_dataset(image_format='np')
        ds_ptformat = df.to_pytorch_dataset(image_format='pt')

        elt_count = 0
        for elt, elt_pt in zip(ds, ds_ptformat):
            arr_plain = elt['c_image']
            assert isinstance(arr_plain, np.ndarray)
            assert arr_plain.flags["WRITEABLE"], 'required by pytorch collate function'

            # NB: compare numpy array bc PIL.Image object itself is not using same file.
            assert (arr_plain == np.array(im_plain)).all(), 'numpy image should be the same as the original'
            arr_xformed = elt['c_image_xformed']
            assert isinstance(arr_xformed, np.ndarray)
            assert arr_xformed.flags["WRITEABLE"], 'required by pytorch collate function'

            assert arr_xformed.shape == (H, W, 3)
            assert arr_xformed.dtype == np.uint8
            # same as above, compare numpy array bc PIL.Image object itself is not using same file.
            assert (arr_xformed == np.array(im_xformed)).all(),\
                'numpy image array for xformed image should be the same as the original'

            # now compare pytorch version
            arr_pt = elt_pt['c_image']
            assert torch.is_tensor(arr_pt)
            arr_pt = elt_pt['c_image_xformed']
            assert torch.is_tensor(arr_pt)
            assert arr_pt.shape == (3, H, W)
            assert arr_pt.dtype == torch.float32
            assert (0.0 <= arr_pt).all()
            assert (arr_pt <= 1.0).all()
            assert torch.isclose(T.ToTensor()(arr_xformed), arr_pt).all(),\
                'pytorch image should be consistent with numpy image'
            elt_count += 1
        assert elt_count == 1

    @pytest.mark.skip('Flaky test (fails intermittently)')
    def test_to_pytorch_dataloader(self, all_datatypes_tbl: catalog.Table) -> None:
        """ Tests the dataset works well with pytorch dataloader:
            1. compatibility with multiprocessing
            2. compatibility of all types with default collate_fn
        """
        import torch.utils.data
        @pxt.udf(param_types=[pxt.JsonType()], return_type=pxt.JsonType())
        def restrict_json_for_default_collate(obj):
            keys = ['id', 'label', 'iscrowd', 'bounding_box']
            return {k: obj[k] for k in keys}
        
        t = all_datatypes_tbl
        df = t.select(
            t.row_id,
            t.c_int,
            t.c_float,
            t.c_bool,
            t.c_timestamp,
            t.c_array,
            t.c_video,
            # default collate_fn doesnt support null values, nor lists of different lengths
            # but does allow some dictionaries if they are uniform
            c_json = restrict_json_for_default_collate(t.c_json.detections[0]),
            # images must be uniform shape for pytorch collate_fn to not fail
            c_image=t.c_image.resize([220, 224]).convert('RGB')
        )
        df_size = df.count()
        ds = df.to_pytorch_dataset(image_format='pt')
        # test serialization:
        #  - pickle.dumps() and pickle.loads() must work so that
        #   we can use num_workers > 0
        x = pickle.dumps(ds)
        _ = pickle.loads(x)

        # test we get all rows
        def check_recover_all_rows(ds, size : int, **kwargs):
            dl = torch.utils.data.DataLoader(ds, **kwargs)
            loaded_ids = set()
            for batch in dl:
                for row_id in batch['row_id']:
                    val = int(row_id) # np.int -> int or will fail set equality test below.
                    assert val not in loaded_ids, val
                    loaded_ids.add(val)

            assert loaded_ids == set(range(size))

        # check different number of workers
        check_recover_all_rows(ds, size=df_size, batch_size=3, num_workers=0) # within this process
        check_recover_all_rows(ds, size=df_size, batch_size=3, num_workers=2) # two separate processes

        # check edge case where some workers get no rows
        short_size = 1
        df_short = df.where(t.row_id < short_size)
        ds_short = df_short.to_pytorch_dataset(image_format='pt')
        check_recover_all_rows(ds_short, size=short_size, batch_size=13, num_workers=short_size+1)

    def test_pytorch_dataset_caching(self, all_datatypes_tbl: catalog.Table) -> None:
        """ Tests that dataset caching works
            1. using the same dataset twice in a row uses the cache
            2. adding a row to the table invalidates the cached version
            3. changing the select list invalidates the cached version
        """
        t = all_datatypes_tbl

        t.drop_column('c_video') # null value video column triggers internal assertions in DataRow
        # see https://github.com/pixeltable/pixeltable/issues/38

        t.drop_column('c_array') # no support yet for null array values in the pytorch dataset

        def _get_mtimes(dir: Path):
            return {p.name: p.stat().st_mtime for p in dir.iterdir()}

        #  check result cached
        ds1 = t.to_pytorch_dataset(image_format='pt')
        ds1_mtimes = _get_mtimes(ds1.path)
        
        ds2 = t.to_pytorch_dataset(image_format='pt')
        ds2_mtimes = _get_mtimes(ds2.path)
        assert ds2.path == ds1.path, 'result should be cached'
        assert ds2_mtimes == ds1_mtimes, 'no extra file system work should have occurred'

        # check invalidation on insert
        t_size = t.count()
        t.insert(row_id=t_size)
        ds3 = t.to_pytorch_dataset(image_format='pt')
        assert ds3.path != ds1.path, 'different path should be used'

        # check select list invalidation
        ds4 = t.select(t.row_id).to_pytorch_dataset(image_format='pt')
        assert ds4.path != ds3.path, 'different select list, hence different path should be used'

    def test_to_coco(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('nos')
        cl = test_client
        base_t = cl.create_table('videos', {'video': pxt.VideoType()})
        args = {'video': base_t.video, 'fps': 1}
        view_t = cl.create_view('frames', base_t, iterator_class=FrameIterator, iterator_args=args)
        from pixeltable.functions.nos.object_detection_2d import yolox_medium
        view_t.add_column(detections=yolox_medium(view_t.frame))
        base_t.insert(video=get_video_files()[0])

        @pxt.udf(return_type=pxt.JsonType(nullable=False), param_types=[pxt.JsonType(nullable=False)])
        def yolo_to_coco(detections):
            bboxes, labels = detections['bboxes'], detections['labels']
            num_annotations = len(detections['bboxes'])
            assert num_annotations == len(detections['labels'])
            result = []
            for i in range(num_annotations):
                bbox = bboxes[i]
                ann = {
                    'bbox': [round(bbox[0]), round(bbox[1]), round(bbox[2] - bbox[0]), round(bbox[3] - bbox[1])],
                    'category': labels[i],
                }
                result.append(ann)
            return result

        query = view_t.select({'image': view_t.frame, 'annotations': yolo_to_coco(view_t.detections)})
        path = query.to_coco_dataset()
        # we get a valid COCO dataset
        coco_ds = COCO(path)
        assert len(coco_ds.imgs) == view_t.count()

        # we call to_coco_dataset() again and get the cached dataset
        new_path = query.to_coco_dataset()
        assert path == new_path

        # the cache is invalidated when we add more data
        base_t.insert(video=get_video_files()[1])
        new_path = query.to_coco_dataset()
        assert path != new_path
        coco_ds = COCO(new_path)
        assert len(coco_ds.imgs) == view_t.count()

        # incorrect select list
        with pytest.raises(excs.Error) as exc_info:
            _ = view_t.select({'image': view_t.frame, 'annotations': view_t.detections}).to_coco_dataset()
        assert '"annotations" is not a list' in str(exc_info.value)

        with pytest.raises(excs.Error) as exc_info:
            _ = view_t.select(view_t.detections).to_coco_dataset()
        assert 'missing key "image"' in str(exc_info.value).lower()



from __future__ import annotations
from typing import Optional, List, Any, Tuple
import io
import urllib

import PIL
import numpy as np


class DataRow:
    """
    Encapsulates all data and execution state needed by RowBuilder and DataRowBatch:
    - state for in-memory computation
    - state for storing the data
    This is not meant to be a black-box abstraction.

    In-memory representations by column type:
    - StringType: str
    - IntType: int
    - FloatType: float
    - BoolType: bool
    - TimestampType: datetime.datetime
    - JsonType: json-serializable object
    - ArrayType: numpy.ndarray
    - ImageType: PIL.Image.Image
    - VideoType: local path if available, otherwise url
    """
    def __init__(self, size: int, img_slot_idxs: List[int], media_slot_idxs: List[int], array_slot_idxs: List[int]):
        self.vals: List[Any] = [None] * size  # either cell values or exceptions
        self.has_val = [False] * size
        self.excs: List[Optional[Exception]] = [None] * size

        # control structures that are shared across all DataRows in a batch
        self.img_slot_idxs = img_slot_idxs
        self.media_slot_idxs = media_slot_idxs  # all media types aside from image
        self.array_slot_idxs = array_slot_idxs

        # the primary key of a store row is a sequence of ints (the number is different for table vs view)
        self.pk: Optional[Tuple[int, ...]] = None

        # file_urls:
        # - stored url of file for media in vals[i]
        # - None if vals[i] is not media type
        # - not None if file_paths[i] is not None
        self.file_urls: List[Optional[str]] = [None] * size

        # file_paths:
        # - local path of media file in vals[i]; points to the file cache if file_urls[i] is remote
        # - None if vals[i] is not a media type or if there is no local file yet for file_urls[i]
        self.file_paths: List[Optional[str]] = [None] * size

    def clear(self) -> None:
        size = len(self.vals)
        self.vals = [None] * size
        self.has_val = [False] * size
        self.excs = [None] * size
        self.pk = None
        self.file_urls = [None] * size
        self.file_paths = [None] * size

    def copy(self, target: DataRow) -> None:
        """Create a copy of the contents of this DataRow in target
        The copy shares the cell values, but not the control structures (eg, self.has_val), because these
        need to be independently updateable.
        """
        target.vals = self.vals.copy()
        target.has_val = self.has_val.copy()
        target.excs = self.excs.copy()
        target.pk = self.pk
        target.file_urls = self.file_urls.copy()
        target.file_paths = self.file_paths.copy()

    def set_pk(self, pk: Tuple[int, ...]) -> None:
        self.pk = pk

    def has_exc(self, slot_idx: int) -> bool:
        return self.excs[slot_idx] is not None

    def get_exc(self, slot_idx: int) -> Exception:
        assert self.has_val[slot_idx] is False
        assert self.excs[slot_idx] is not None
        return self.excs[slot_idx]

    def set_exc(self, slot_idx: int, exc: Exception) -> None:
        assert self.excs[slot_idx] is None
        self.excs[slot_idx] = exc

        if self.has_val[slot_idx]:
            # eg. during validation, where contents of file is found invalid
            self.has_val[slot_idx] = False
            self.vals[slot_idx] = None
            self.file_paths[slot_idx] = None
            self.file_urls[slot_idx] = None

    def __getitem__(self, index: object) -> Any:
        """Returns in-memory value, ie, what is needed for expr evaluation"""
        if not self.has_val[index]:
            # for debugging purposes
            pass
        assert self.has_val[index], index

        if self.file_urls[index] is not None and index in self.img_slot_idxs:
            # if we need to load this from a file, it should have been materialized locally
            assert self.file_paths[index] is not None
            if self.vals[index] is None:
                self.vals[index] = PIL.Image.open(self.file_paths[index])

        return self.vals[index]

    def get_stored_val(self, index: object) -> Any:
        """Return the value that gets stored in the db"""
        assert self.excs[index] is None
        if not self.has_val[index]:
            # for debugging purposes
            pass
        assert self.has_val[index]

        if self.file_urls[index] is not None and (index in self.img_slot_idxs or index in self.media_slot_idxs):
            # if this is an image or other media type we want to store, we should have a url
            return self.file_urls[index]

        if self.vals[index] is not None and index in self.array_slot_idxs:
            assert isinstance(self.vals[index], np.ndarray)
            np_array = self.vals[index]
            buffer = io.BytesIO()
            np.save(buffer, np_array)
            return buffer.getvalue()

        return self.vals[index]

    def __setitem__(self, idx: object, val: Any) -> None:
        """Assign in-memory cell value
        This allows overwriting
        """
        assert self.excs[idx] is None

        if (idx in self.img_slot_idxs or idx in self.media_slot_idxs) and isinstance(val, str):
            # this is either a local file path or a URL
            parsed = urllib.parse.urlparse(val)
            if parsed.scheme == '' or parsed.scheme == 'file':
                # local file path
                assert self.file_urls[idx] is None and self.file_paths[idx] is None
                if parsed.scheme == '':
                    self.file_urls[idx] = urllib.parse.urljoin('file:', urllib.request.pathname2url(parsed.path))
                else:
                    self.file_urls[idx] = val
                self.file_paths[idx] = urllib.parse.unquote(parsed.path)
            else:
                # URL
                assert self.file_urls[idx] is None
                self.file_urls[idx] = val

            if idx in self.media_slot_idxs:
                self.vals[idx] = self.file_paths[idx] if self.file_paths[idx] is not None else self.file_urls[idx]
        elif idx in self.array_slot_idxs and isinstance(val, bytes):
            self.vals[idx] = np.load(io.BytesIO(val))
        else:
            self.vals[idx] = val
        self.has_val[idx] = True

    def set_file_path(self, idx: object, path: str) -> None:
        """Augment an existing url with a local file path"""
        assert self.has_val[idx]
        assert idx in self.img_slot_idxs or idx in self.media_slot_idxs
        self.file_paths[idx] = path
        if idx in self.media_slot_idxs:
            self.vals[idx] = path

    def flush_img(self, index: object, filepath: Optional[str] = None) -> None:
        """Discard the in-memory value and save it to a local file, if filepath is not None"""
        if self.vals[index] is None:
            return
        assert self.excs[index] is None
        if self.file_paths[index] is None:
            if filepath is not None:
                # we want to save this to a file
                self.file_paths[index] = filepath
                self.file_urls[index] = urllib.parse.urljoin('file:', urllib.request.pathname2url(filepath))
                self.vals[index].save(filepath, format='JPEG')
            else:
                # we discard the content of this cell
                self.has_val[index] = False
        else:
            # we already have a file for this image, nothing left to do
            pass
        self.vals[index] = None




from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple

import sqlalchemy as sql

from .globals import ComparisonOperator
from .expr import Expr
from .predicate import Predicate
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog


class Comparison(Predicate):
    def __init__(self, operator: ComparisonOperator, op1: Expr, op2: Expr):
        super().__init__()
        self.operator = operator
        self.components = [op1, op2]
        self.id = self._create_id()

    def __str__(self) -> str:
        return f'{self._op1} {self.operator} {self._op2}'

    def _equals(self, other: Comparison) -> bool:
        return self.operator == other.operator

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('operator', self.operator.value)]

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        left = self._op1.sql_expr()
        right = self._op2.sql_expr()
        if left is None or right is None:
            return None
        if self.operator == ComparisonOperator.LT:
            return left < right
        if self.operator == ComparisonOperator.LE:
            return left <= right
        if self.operator == ComparisonOperator.EQ:
            return left == right
        if self.operator == ComparisonOperator.NE:
            return left != right
        if self.operator == ComparisonOperator.GT:
            return left > right
        if self.operator == ComparisonOperator.GE:
            return left >= right

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        if self.operator == ComparisonOperator.LT:
            data_row[self.slot_idx] = data_row[self._op1.slot_idx] < data_row[self._op2.slot_idx]
        elif self.operator == ComparisonOperator.LE:
            data_row[self.slot_idx] = data_row[self._op1.slot_idx] <= data_row[self._op2.slot_idx]
        elif self.operator == ComparisonOperator.EQ:
            data_row[self.slot_idx] = data_row[self._op1.slot_idx] == data_row[self._op2.slot_idx]
        elif self.operator == ComparisonOperator.NE:
            data_row[self.slot_idx] = data_row[self._op1.slot_idx] != data_row[self._op2.slot_idx]
        elif self.operator == ComparisonOperator.GT:
            data_row[self.slot_idx] = data_row[self._op1.slot_idx] > data_row[self._op2.slot_idx]
        elif self.operator == ComparisonOperator.GE:
            data_row[self.slot_idx] = data_row[self._op1.slot_idx] >= data_row[self._op2.slot_idx]

    def _as_dict(self) -> Dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'operator' in d
        return cls(ComparisonOperator(d['operator']), components[0], components[1])

